import asyncio
import logging
from datetime import datetime, timedelta
from decimal import Decimal
from typing import TYPE_CHECKING, Optional, Union

from ya_payment import RequestorApi, models

from golem.resources.allocation import Allocation
from golem.resources.base import _NULL, Resource, api_call_wrapper
from golem.resources.debit_note.events import NewDebitNote
from golem.resources.exceptions import PaymentValidationException
from golem.resources.utils.infrastructure import InfrastructureProps
from golem.resources.utils.payment import (
    LinearCoeffs,
    PayDocumentStatus,
    PaymentProps,
    eth_decimal,
    validate_payment_calculated_cost,
    validate_payment_max_cost,
)

if TYPE_CHECKING:
    from golem.node import GolemNode
    from golem.resources.activity import Activity  # noqa
    from golem.resources.agreement.data import AgreementData  # noqa

logger = logging.getLogger(__name__)


class DebitNote(Resource[RequestorApi, models.DebitNote, "Activity", _NULL, _NULL]):
    """A single debit note on the Golem Network.

    Usually created by a :any:`GolemNode` initialized with `collect_payment_events = True`.
    """

    def __init__(self, node: "GolemNode", id_: str, data: Optional[models.DebitNote] = None):
        super().__init__(node, id_, data)
        asyncio.create_task(node.event_bus.emit(NewDebitNote(self)))

    @property
    def activity(self) -> "Activity":
        return self.parent

    async def get_status(self) -> PayDocumentStatus:
        return PayDocumentStatus(str((await self.get_data()).status).upper())

    async def get_previous_debit_note(self) -> Optional["DebitNote"]:
        """Get previous debit note."""
        return max(
            (dn for dn in self.activity.debit_notes if dn.created_at < self.created_at),
            key=lambda dn: dn.created_at,  # type: ignore[union-attr]
            default=None,
        )

    async def get_previous_debit_notes_count(self) -> int:
        """Get previous debit notes count."""
        return len([dn for dn in self.activity.debit_notes if dn.created_at < self.created_at])

    async def get_previous_payable_debit_notes_count(self) -> int:
        """Get previous payable debit notes count."""
        return len(
            [
                dn
                for dn in self.activity.debit_notes
                if dn.created_at < self.created_at
                and (await dn.get_data()).payment_due_date is not None
            ]
        )

    @staticmethod
    def validate_mid_agreement_payment(
        activity_created_at: datetime,
        debit_note_created_at: datetime,
        payment_props: PaymentProps,
        payment_due_date: Optional[datetime],
        previous_debit_notes_count: int = 0,
        previous_payable_debit_notes_count: int = 0,
        payment_timeout_grace_period: timedelta = timedelta(minutes=5),
        debit_note_interval_grace_period: timedelta = timedelta(seconds=30),
    ) -> None:
        """Validate debit note mid agreement payment data."""
        if payment_due_date is None:
            return

        if payment_props.payment_timeout is None or payment_props.debit_note_interval is None:
            raise PaymentValidationException(
                "Payable debit note received when mid-agreement payments inactive."
            )

        payment_timeout_timedelta = timedelta(seconds=payment_props.payment_timeout)

        received_payment_timeout = payment_due_date - debit_note_created_at
        if received_payment_timeout + payment_timeout_grace_period < payment_timeout_timedelta:
            raise PaymentValidationException(
                f"Payment timeout is shorter than agreed {payment_due_date=}"
                f"{received_payment_timeout=} < {payment_timeout_timedelta=}."
            )

        debit_note_interval = timedelta(seconds=payment_props.debit_note_interval)
        activity_duration = debit_note_created_at - activity_created_at
        if (
            activity_duration + debit_note_interval_grace_period
            < (previous_debit_notes_count - 1) * debit_note_interval
        ):
            raise PaymentValidationException(
                f"Too many debit notes received {previous_debit_notes_count=}. "
                f"{activity_duration + debit_note_interval_grace_period}"
                f"< {previous_debit_notes_count * debit_note_interval}"
            )

        payable_debit_note_interval = timedelta(seconds=payment_props.payment_timeout)
        activity_duration = debit_note_created_at - activity_created_at
        if (
            activity_duration + payment_timeout_grace_period
            < (previous_payable_debit_notes_count - 1) * payable_debit_note_interval
        ):
            raise PaymentValidationException(
                f"Too many payable debit notes received {previous_payable_debit_notes_count=}."
                f"{activity_duration + payment_timeout_grace_period}"
                f"< {previous_payable_debit_notes_count * payable_debit_note_interval}"
            )

    def validate_payment_data(
        self,
        debit_note_data: models.DebitNote,
        previous_debit_note: Optional["DebitNote"],
        previous_debit_notes_count: int,
        previous_payable_debit_notes_count: int,
        agreement_data: "AgreementData",
    ) -> Decimal:
        """Validate debit note payment data.

        Raises: PaymentValidationException
        """
        if agreement_data.agreement_duration is None:
            raise PaymentValidationException("Agreement was not approved")

        payment_props = PaymentProps.from_properties(agreement_data.properties)
        self.validate_mid_agreement_payment(
            self.activity.created_at,
            self.created_at,
            payment_props,
            debit_note_data.payment_due_date,
            previous_debit_notes_count,
            previous_payable_debit_notes_count,
        )

        coeffs = LinearCoeffs.from_properties(agreement_data.properties)
        if coeffs is None:
            raise PaymentValidationException("Unable to retrieve coeffs details")

        total_amount_due = eth_decimal(debit_note_data.total_amount_due)
        calculated_cost = validate_payment_calculated_cost(
            coeffs=coeffs,
            amount=total_amount_due,
            usage_counter_vector=debit_note_data.usage_counter_vector,  # type: ignore
        )

        time_since_last_dn = amount_since_last_dn = None
        if previous_debit_note:
            time_since_last_dn = self.created_at - previous_debit_note.created_at
            amount_since_last_dn = total_amount_due - eth_decimal(
                previous_debit_note.data.total_amount_due
            )

        max_cost, max_cost_since_last_debit_note = validate_payment_max_cost(
            coeffs=coeffs,
            inf=InfrastructureProps.from_properties(agreement_data.properties),
            duration=self.created_at - self.activity.created_at,
            amount=total_amount_due,
            time_since_last_debit_note=time_since_last_dn,
            amount_since_last_debit_note=amount_since_last_dn,
        )

        logger.info(
            f"Validation of DebitNote for {agreement_data.agreement_id=} ok "
            f"{total_amount_due=} {calculated_cost=} {max_cost=} {max_cost_since_last_debit_note=}"
        )
        return total_amount_due

    async def validate_and_accept(self, allocation: Allocation) -> None:
        """Validate debit note and accept using a given :any:`Allocation`."""
        debit_note_data = await self.get_data(force=True)
        if debit_note_data.status != PayDocumentStatus.RECEIVED:
            logger.warning(f"Wrong status of debit_note {debit_note_data.status} != RECEIVED")
            return

        agreement_data: "AgreementData" = await self.activity.agreement.get_agreement_data()

        previous_debit_note = await self.get_previous_debit_note()
        previous_debit_notes_count = await self.get_previous_debit_notes_count()
        previous_payable_debit_notes_count = await self.get_previous_payable_debit_notes_count()

        try:
            total_amount_due = self.validate_payment_data(
                debit_note_data,
                previous_debit_note,
                previous_debit_notes_count,
                previous_payable_debit_notes_count,
                agreement_data,
            )
        except PaymentValidationException:
            logger.warning(f"Debit Note {self.id} validation failed", exc_info=True)
            return

        await self.accept(allocation, total_amount_due)

    async def accept_full(self, allocation: Allocation) -> None:
        """Accept full debit note amount using a given :any:`Allocation`."""
        amount_str = (await self.get_data()).total_amount_due
        await self.accept(allocation, Decimal(amount_str))

    @api_call_wrapper(retry_count=5)
    async def accept(self, allocation: Allocation, amount: Union[Decimal, float]) -> None:
        acceptance = models.Acceptance(
            total_amount_accepted=str(amount), allocation_id=allocation.id
        )
        await self.api.accept_debit_note(self.id, acceptance)



from copy import deepcopy
from typing import Any, Dict

from golem.payload.mixins import PropsConsSerializerMixin

_missing = object()


class Properties(PropsConsSerializerMixin, dict):
    """Low level wrapper class for Golem's Market API properties manipulation."""

    def __init__(self, mapping=_missing, /) -> None:
        if mapping is _missing:
            super().__init__()
            return

        mapping_deep_copy = deepcopy(mapping)

        super().__init__(mapping_deep_copy)

    def serialize(self) -> Dict[str, Any]:
        """Serialize complex objects into format handled by Market API properties specification."""
        return {
            key: self._serialize_property(value) for key, value in self.items() if value is not None
        }

    def _serialize_property(self, value: Any) -> Any:
        return self._serialize_value(value)



import asyncio
import logging
from copy import deepcopy
from datetime import timedelta
from typing import Optional, Sequence

from ya_market import ApiException

from golem.managers import ProposalManagerPlugin, RejectProposal
from golem.managers.base import ManagerPluginException, ProposalNegotiator
from golem.resources import DemandData, Proposal
from golem.resources.proposal.exceptions import ProposalRejected
from golem.utils.asyncio.tasks import resolve_maybe_awaitable
from golem.utils.logging import trace_span

logger = logging.getLogger(__name__)

DEFAULT_PROPOSAL_RESPONSE_TIMEOUT = timedelta(seconds=5)


class NegotiatingPlugin(ProposalManagerPlugin):
    def __init__(
        self,
        proposal_negotiators: Optional[Sequence[ProposalNegotiator]] = None,
        proposal_response_timeout: timedelta = DEFAULT_PROPOSAL_RESPONSE_TIMEOUT,
        *args,
        **kwargs,
    ) -> None:
        self._proposal_negotiators: Sequence[ProposalNegotiator] = (
            list(proposal_negotiators) if proposal_negotiators is not None else []
        )
        self._proposal_response_timeout = proposal_response_timeout.total_seconds()

        self._success_count = 0
        self._fail_count = 0

        super().__init__(*args, **kwargs)

    @trace_span(show_results=True)
    async def get_proposal(self) -> Proposal:
        while True:
            proposal = await self._get_proposal()
            demand_data = await proposal.demand.get_demand_data()
            provider_name = await proposal.get_provider_name()

            try:
                negotiated_proposal = await self._negotiate_proposal(demand_data, proposal)
            except Exception as e:
                self._fail_count += 1
                if isinstance(e, ProposalRejected):
                    reason = f" as it was rejected by the provider: `{e}`"
                elif isinstance(e, RejectProposal):
                    reason = f" as it was rejected by the requestor: `{e}`"
                else:
                    reason = ""

                logger.debug(
                    f"Negotiation based on proposal `{proposal}` from `{provider_name}`"
                    f" failed{reason}, retrying with new one..."
                    "\nsuccess count: "
                    f"{self._success_count}/{self._success_count + self._fail_count}",
                    exc_info=not isinstance(e, (ProposalRejected, RejectProposal)),
                )
            else:
                self._success_count += 1
                logger.info(
                    f"Negotiation based on proposal `{proposal}` from `{provider_name}` succeeded"
                    "\nsuccess count: "
                    f"{self._success_count}/{self._success_count + self._fail_count}"
                )
                return negotiated_proposal

    @trace_span(show_arguments=True, show_results=True)
    async def _negotiate_proposal(
        self, demand_data: DemandData, offer_proposal: Proposal
    ) -> Proposal:
        while True:
            demand_data_after_negotiators = deepcopy(demand_data)

            try:
                await self._run_negotiators(demand_data_after_negotiators, offer_proposal)
            except RejectProposal:
                if not offer_proposal.initial:
                    await self._reject_proposal(offer_proposal)

                raise

            if not offer_proposal.initial and demand_data_after_negotiators == demand_data:
                return offer_proposal

            demand_data = demand_data_after_negotiators

            demand_proposal = await self._send_demand_proposal(offer_proposal, demand_data)
            new_offer_proposal = await self._wait_for_proposal_response(demand_proposal)

            offer_proposal = new_offer_proposal

    @trace_span()
    async def _wait_for_proposal_response(self, demand_proposal: Proposal) -> Proposal:
        try:
            return await asyncio.wait_for(
                demand_proposal.responses().__anext__(),
                timeout=self._proposal_response_timeout,
            )
        except (StopAsyncIteration, asyncio.TimeoutError) as e:
            raise ManagerPluginException("Failed to receive proposal response!") from e

    @trace_span()
    async def _send_demand_proposal(
        self, offer_proposal: Proposal, demand_data: DemandData
    ) -> Proposal:
        try:
            return await offer_proposal.respond(
                demand_data.properties,
                demand_data.constraints,
            )
        except (ApiException, asyncio.TimeoutError) as e:
            raise ManagerPluginException(f"Failed to send proposal response! {e}") from e

    @trace_span()
    async def _reject_proposal(self, offer_proposal: Proposal) -> None:
        await offer_proposal.reject()

    @trace_span()
    async def _run_negotiators(
        self, demand_data_after_negotiators: DemandData, offer_proposal: Proposal
    ):
        proposal_data = await offer_proposal.get_proposal_data()

        for negotiator in self._proposal_negotiators:
            negotiator_result = await resolve_maybe_awaitable(
                negotiator(demand_data_after_negotiators, proposal_data)
            )

            if isinstance(negotiator_result, RejectProposal):
                raise negotiator_result

            # Note: Explicit identity to False desired here, not "falsy" check
            if negotiator_result is False:
                raise RejectProposal()



from typing import TYPE_CHECKING

from golem.resources.events import NewResource, ResourceClosed, ResourceDataChanged

if TYPE_CHECKING:
    from golem.resources.debit_note.debit_note import DebitNote  # noqa


class NewDebitNote(NewResource["DebitNote"]):
    pass


class DebitNoteDataChanged(ResourceDataChanged["DebitNote"]):
    pass


class DebitNoteClosed(ResourceClosed["DebitNote"]):
    pass



"""Golem File Transfer Storage Provider."""

import asyncio
import contextlib
import json
import logging
import os
import sys
import tempfile
from dataclasses import dataclass
from os import PathLike
from pathlib import Path
from types import TracebackType
from typing import (
    AsyncContextManager,
    AsyncIterator,
    Dict,
    Iterator,
    List,
    Optional,
    Set,
    Type,
    Union,
    cast,
)

import jsonrpc_base
import semantic_version
from async_exit_stack import AsyncExitStack
from typing_extensions import Literal, Protocol, TypedDict

from golem.utils.storage.base import Content, Destination, Source, StorageProvider
from golem.utils.storage.utils import strtobool

_logger = logging.getLogger(__name__)


class PubLink(TypedDict):
    """GFTP linking information."""

    file: str
    """file on local filesystem."""

    url: str
    """GFTP url as which local files is exposed."""


CommandStatus = Literal["ok", "error"]


class GftpDriver(Protocol):
    """Golem FTP service API."""

    async def version(self) -> str:
        """Get driver version."""

    async def publish(self, *, files: List[str]) -> List[PubLink]:
        """Expose local file as GFTP url.

        `files`
        :   local files to be exposed

        """

    async def close(self, *, urls: List[str]) -> List[CommandStatus]:
        """Stop exposing GFTP urls created by [publish(files=[..])](#publish)."""

    async def receive(self, *, output_file: str) -> PubLink:
        """Create GFTP url for receiving file.

        :  `output_file` -
        """

    async def upload(self, *, file: str, url: str) -> None:
        pass

    async def shutdown(self) -> CommandStatus:
        """Stop GFTP service.

        After shutdown all generated urls will be unavailable.
        """


def service(debug: bool = False) -> AsyncContextManager[GftpDriver]:
    proc = __Process(_debug=debug)
    return cast(AsyncContextManager[GftpDriver], proc)


class __Process(jsonrpc_base.Server):
    def __init__(self, _debug: bool = False):
        super().__init__()
        self._debug = _debug
        self._proc: Optional[asyncio.subprocess.Process] = None
        self._lock = asyncio.Lock()

    async def __aenter__(self) -> GftpDriver:
        env = dict(os.environ, RUST_LOG="debug") if self._debug else None
        self._proc = await asyncio.create_subprocess_shell(
            "gftp server",
            stdout=asyncio.subprocess.PIPE,
            stdin=asyncio.subprocess.PIPE,
            env=env,
        )
        return cast(GftpDriver, self)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        with contextlib.suppress(Exception):
            await self._close()

    async def _close(self) -> None:
        if self._proc is None:
            return
        p: asyncio.subprocess.Process = self._proc
        self._proc = None

        with contextlib.suppress(Exception):
            await cast(GftpDriver, self).shutdown()

        async with self._lock:
            if p.stdin:
                await p.stdin.drain()
                p.stdin.close()
                try:
                    await asyncio.wait_for(p.wait(), 10.0)
                    return
                except asyncio.TimeoutError:
                    pass
        p.kill()
        ret_code = await p.wait()
        _logger.debug("GFTP server closed, code=%d", ret_code)

    def __log_debug(self, msg_dir: Literal["in", "out"], msg: Union[bytes, str]) -> None:
        if self._debug:
            if isinstance(msg, bytes):
                msg = msg.decode(encoding="utf-8")
            stderr = sys.stderr
            stderr.write("\n <= " if msg_dir == "in" else "\n => ")
            stderr.write(msg)
            stderr.flush()

    async def send_message(self, message):
        async with self._lock:
            assert self._proc is not None
            assert self._proc.stdin is not None
            assert self._proc.stdout is not None
            bytes = message.serialize() + "\n"
            self.__log_debug("out", bytes)
            self._proc.stdin.write(bytes.encode("utf-8"))
            await self._proc.stdin.drain()
            msg = await self._proc.stdout.readline()
            self.__log_debug("in", msg)
            if not msg:
                sys.stderr.write("Please check if gftp is installed and is in your $PATH.\n")
                sys.stderr.flush()
            msg = json.loads(msg)
            return message.parse_response(msg)


class GftpSource(Source):
    def __init__(self, length: int, link: PubLink):
        self._len = length
        self._link = link

    @property
    def download_url(self) -> str:
        return self._link["url"]

    @property
    def path(self) -> Path:
        return Path(self._link["file"])

    async def content_length(self) -> int:
        return self._len


class GftpDestination(Destination):
    def __init__(self, _proc: GftpDriver, _link: PubLink) -> None:
        self._proc = _proc
        self._link = _link

    @property
    def upload_url(self) -> str:
        return self._link["url"]

    async def download_stream(self) -> Content:
        file_path = Path(self._link["file"])
        length = file_path.stat().st_size

        async def chunks() -> AsyncIterator[bytes]:
            with open(file_path, "rb") as f:
                chunk = f.read(30_000)
                while chunk:
                    yield chunk
                    chunk = f.read(30_000)

        return Content(length=length, stream=chunks())

    async def download_file(self, destination_file: PathLike) -> None:
        if str(destination_file) == self._link["file"]:
            return
        return await super().download_file(destination_file)


@contextlib.contextmanager
def _temp_file(temp_dir: Path) -> Iterator[Path]:
    """Create a new temporary file in `temp_dir`.

    Implements the ContextManager interface. Deletes the file on ContextManager's exit.
    """
    fd, name = tempfile.mkstemp(prefix="storage_gftp_", dir=temp_dir)
    os.close(fd)
    path = Path(name)
    try:
        yield path
    finally:
        _delete_if_exists(path)


def _delete_if_exists(path: Path) -> None:
    if path.exists():
        try:
            path.unlink()
            _logger.debug("Deleted temporary file %s", path)
        except PermissionError as err:
            # We're on Windows and using `gftp` < 0.7.3, so the file is kept open
            # by `gftp` and cannot be deleted.
            _logger.debug("Cannot delete file: %s", err)


USE_GFTP_CLOSE_ENV_VAR = "YAPAPI_USE_GFTP_CLOSE"
"""The environment variable used by GftpProvider to control whether `gftp close` should be used."""


def read_use_gftp_close_env_var() -> Optional[bool]:
    """Determine from the environment whether `GftpProvider` should use the `gftp close` command.

    Reads the value of the environment variable with the name stored in `USE_GFTP_CLOSE_ENV_VAR`.
    If the environment variable is set and its value can be interpreted as boolean value by
    `distutils.util.strtobool()` then the corresponding boolean value is returned. Otherwise,
    returns `None`.
    """
    try:
        env_value = os.environ.get(USE_GFTP_CLOSE_ENV_VAR, "")
        return bool(strtobool(env_value))
    except ValueError:
        return None


MIN_GFTP_VERSION_THAT_CAN_GFTP_CLOSE = semantic_version.Version("0.7.3")


class GftpProvider(StorageProvider, AsyncContextManager[StorageProvider]):
    """A StorageProvider that communicates with `gftp server` through JSON-RPC.

    The provider keeps track of the files published by `gftp` and their URLs.
    If an URL no longer needs to be published then the provider _should_ issue the
    `gftp close URL` command, so the file published with this URL is closed by `gftp`.

    However, `gftp close URL` may cause errors, due to a bug in `gftp` prior to version 0.7.3
    (see https://github.com/golemfactory/yagna/pull/1501). Therefore the provider uses
    the following logic to dermine if it should use the `gftp close URL` command:
    1. If the environment variable `YAPAPI_USE_GFTP_CLOSE` is set to a truthy value,
       then `gftp close URL` will be used.
    2. If the environment variable `YAPAPI_USE_GFTP_CLOSE` is set to a falsy value,
       then `gftp close URL` will not be used.
    3. If neither 1 nor 2 holds and the version reported by `gftp` is 0.7.3 or larger
       (according to Semantic Versioning 2.0.0) then `gftp close URL` will be used.
    4. Otherwise `gftp close URL` will not be used.

    Note: Reading the `YAPAPI_USE_GFTP_CLOSE` variable is done once, when the provider
    is instantiated, and the version check is made in the provider's `__aenter__()` method.
    """

    @dataclass
    class URLInfo:
        """Information about an URL published through `gftp`."""

        publish_count: int
        """Number of `gftp publish` operations for this URL.

        Serves as a reference counter. When it drops to 0, `gftp close {URL}` is invoked
        in order to release any file published with this URL that is kept open by `gftp`.
        Note that the value of this field may be larger than the number of files published,
        since a single file may be published more than once."""

        temporary_files: Set[Path]
        """Set of temporary files published with this URL.

        When the URL is unpublished by calling `gftp close {URL}`, all temporary files with this
        URL can be safely deleted.
        """

    def __init__(self, *, tmpdir: Optional[str] = None):
        self.__exit_stack = AsyncExitStack()

        # Directory for temporal files created by this provider
        self._temp_dir: Optional[Path] = Path(tmpdir) if tmpdir else None

        # Mapping of URLs to info on files published with this URL
        self._published_sources: Dict[str, GftpProvider.URLInfo] = dict()

        # Lock used to synchronize access to self._published_sources
        self._lock: asyncio.Lock = asyncio.Lock()

        # Flag indicating if this `GftpProvider` will close unpublished URLs.
        # See this class' docstring for more details.
        self._close_urls: Optional[bool] = read_use_gftp_close_env_var()

        # Reference to an external process running the `gftp server` command
        self._process: Optional["__Process"] = None

    async def __aenter__(self) -> StorageProvider:
        if not self._temp_dir:
            self._temp_dir = Path(
                self.__exit_stack.enter_context(tempfile.TemporaryDirectory(prefix="sdk-gftp-"))
            )
            _logger.debug("Creating a temporary directory %s", self._temp_dir)
        process = await self.__get_process()
        gftp_version = await process.version()
        assert gftp_version

        if self._close_urls is None:
            try:
                # Gftp_version could be something like `7.2.3 (10116c7d 2021-07-28 build #164)`,
                # we need to discard everything after the first space.
                semver = semantic_version.Version(gftp_version.split()[0])
                self._close_urls = semver >= MIN_GFTP_VERSION_THAT_CAN_GFTP_CLOSE
                _logger.debug(
                    "Setting _close_urls to %s, gftp version: %s",
                    self._close_urls,
                    gftp_version,
                )
            except ValueError:
                _logger.warning("Cannot parse gftp version info '%s'", gftp_version)
                self._close_urls = False
        assert self._close_urls is not None

        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        await self.__exit_stack.aclose()
        # Remove temporary files created by this provider
        if not self._temp_dir:
            raise RuntimeError("GftpProvider.__aenter__() not called")
        if self._temp_dir and self._temp_dir.exists():
            for info in self._published_sources.values():
                for path in info.temporary_files:
                    _delete_if_exists(path)

        return None

    def __new_file(self) -> Path:
        if not self._temp_dir:
            raise RuntimeError("GftpProvider.__aenter__() not called")
        return self.__exit_stack.enter_context(_temp_file(self._temp_dir))

    async def __get_process(self) -> GftpDriver:
        _debug = bool(os.getenv("DEBUG_GFTP"))
        process = self._process or (await self.__exit_stack.enter_async_context(service(_debug)))
        if not self._process:
            self._process = process
        return process

    async def upload_stream(self, length: int, stream: AsyncIterator[bytes]) -> Source:
        file_name = self.__new_file()
        with open(file_name, "wb") as f:
            async for chunk in stream:
                f.write(chunk)
        return await self.upload_file(file_name, _temporary=True)

    async def upload_file(self, path: os.PathLike, _temporary: bool = False) -> Source:
        path = Path(path)
        _logger.debug("Publishing file %s...", path)
        process = await self.__get_process()

        async with self._lock:
            links = await process.publish(files=[str(path)])
            assert len(links) == 1, "Invalid gftp publish response"

            length = path.stat().st_size

            url = links[0]["url"]

            if url not in self._published_sources:
                info = GftpProvider.URLInfo(
                    publish_count=1,
                    temporary_files=({path} if _temporary else set()),
                )
                self._published_sources[url] = info
            else:
                info = self._published_sources[url]

                if path in info.temporary_files:
                    raise ValueError(f"File {path} already published as temporary")

                if _temporary:
                    info.temporary_files.add(path)
                info.publish_count += 1

            _logger.debug(
                "File %s published with URL = %s, count = %d",
                path,
                url,
                info.publish_count,
            )

        source = GftpSource(length, links[0])
        return source

    async def release_source(self, source: Source) -> None:
        if not isinstance(source, GftpSource):
            raise ValueError(f"Expected an instance of GftpSource, got {type(source)} instead")

        url = source.download_url
        _logger.debug("Releasing file %s with URL = %s ...", source.path, url)

        async with self._lock:
            if url not in self._published_sources:
                raise ValueError(
                    f"Trying to release an unpublished URL {url}, path = {source.path}"
                )
            info = self._published_sources[url]
            info.publish_count -= 1

            _logger.debug(
                "File %s released, URL = %s, count = %d",
                source.path,
                url,
                info.publish_count,
            )

            if info.publish_count == 0:
                _logger.debug("Unpublishing URL %s...", url)
                if self._close_urls:
                    process = await self.__get_process()
                    await process.close(urls=[url])

                for path in info.temporary_files:
                    _delete_if_exists(path)

                del self._published_sources[url]

    async def new_destination(self, destination_file: Optional[PathLike] = None) -> Destination:
        if destination_file:
            if Path(destination_file).exists():
                destination_file = None
        output_file = str(destination_file) if destination_file else str(self.__new_file())
        process = await self.__get_process()
        link = await process.receive(output_file=output_file)
        return GftpDestination(process, link)


def provider() -> AsyncContextManager[StorageProvider]:
    return GftpProvider()


__all__ = ("service", "provider", "GftpDriver", "PubLink")



import asyncio
import logging
from typing import Generic, List, Optional, Sequence

from golem.managers.base import ManagerException, TPlugin
from golem.utils.asyncio import create_task_with_logging, ensure_cancelled
from golem.utils.logging import get_trace_id_name, trace_span

logger = logging.getLogger(__name__)


class BackgroundLoopMixin:
    def __init__(self, *args, **kwargs) -> None:
        self._background_loop_task: Optional[asyncio.Task] = None

        super().__init__(*args, **kwargs)

    async def start(self) -> None:
        if self.is_started():
            raise ManagerException("Already started!")

        self._background_loop_task = create_task_with_logging(
            self._background_loop(),
            trace_id=get_trace_id_name(self, "background-loop"),
        )

    async def stop(self) -> None:
        if self._background_loop_task is not None:
            await ensure_cancelled(self._background_loop_task)
            self._background_loop_task = None

    def is_started(self) -> bool:
        return self._background_loop_task is not None and not self._background_loop_task.done()

    async def _background_loop(self) -> None:
        pass


class PluginsMixin(Generic[TPlugin]):
    def __init__(self, plugins: Optional[Sequence[TPlugin]] = None, *args, **kwargs) -> None:
        self._plugins: List[TPlugin] = list(plugins) if plugins is not None else []

        super().__init__(*args, **kwargs)

    @trace_span()
    def register_plugin(self, plugin: TPlugin):
        self._plugins.append(plugin)

    @trace_span()
    def unregister_plugin(self, plugin: TPlugin):
        self._plugins.remove(plugin)



import asyncio
import inspect
from typing import AsyncIterator, Awaitable, Generic, List, TypeVar, Union

from golem.pipeline.exceptions import InputStreamExhausted

DataType = TypeVar("DataType")


class Buffer(Generic[DataType]):
    """
    Process concurrently multiple elements from an async stream of awaitables.

    Sample usage::

        async def awaitable_foo():
            await asyncio.sleep(1)
            return "foo"

        async def awaitable_bar():
            await asyncio.sleep(1)
            return "bar"

        async def stream():
            yield awaitable_foo()
            yield awaitable_bar()

        #   Prints "foo" after a second and "bar" after another second
        async for x in Buffer()(stream()):
            print(x)

        #   Prints "foo" and "bar" after a single second
        async for x in Buffer(size=2)(stream()):
            print(x)

    """

    def __init__(self, size: int = 1):
        """Init Buffer.

        :param size: How many elements of the input stream will be concurrently awaited.
            Default size=1 is identical to a single `await` statement.
            In most Golem-specific scenarios buffer size will correspond to things like
            "how many agreements are negotiated at the same time", so usually the higher the size:

            * The faster we'll be able to create and utilize resources
            * The higher chance for "useless" resources (e.g. we might be creating multiple
              agreements even when there is only a single task left)
            * The higher workload for the local machine (more asyncio tasks) and yagna
              (more requests)
        """
        self.size = size

    async def __call__(
        self, in_stream: AsyncIterator[Union[DataType, Awaitable[DataType]]]
    ) -> AsyncIterator[DataType]:
        """Call Buffer.

        :param in_stream: A stream of awaitables.
        """
        self._tasks: List[asyncio.Task] = []
        self._in_stream_exhausted = False
        self._result_queue: asyncio.Queue[DataType] = asyncio.Queue()
        self._main_task = asyncio.create_task(self._process_in_stream(in_stream))
        self._semaphore = asyncio.BoundedSemaphore(self.size)

        stop_task = asyncio.create_task(self._wait_until_empty())

        while True:
            get_result_task = asyncio.create_task(self._result_queue.get())
            await asyncio.wait((get_result_task, stop_task), return_when=asyncio.FIRST_COMPLETED)
            if stop_task.done():
                get_result_task.cancel()
                break
            else:
                yield get_result_task.result()
                self._semaphore.release()

    async def _process_in_stream(
        self, in_stream: AsyncIterator[Union[DataType, Awaitable[DataType]]]
    ) -> None:
        while True:
            await self._semaphore.acquire()
            try:
                in_val = await in_stream.__anext__()
            except StopAsyncIteration:
                self._semaphore.release()
                self._in_stream_exhausted = True
                return
            task = asyncio.create_task(self._process_single_value(in_val))
            self._tasks.append(task)

    async def _process_single_value(self, in_val: Union[DataType, Awaitable[DataType]]) -> None:
        awaited: DataType
        if inspect.isawaitable(in_val):
            try:
                awaited = await in_val
            except InputStreamExhausted:
                self._main_task.cancel()
                self._semaphore.release()
                self._in_stream_exhausted = True
                return
            except Exception as e:
                self._semaphore.release()
                #   TODO https://github.com/golemfactory/golem-core-python/issues/27
                print("Exception in Buffer", e)
                return
        else:
            #   NOTE: Buffer is useful only with awaitables, so this scenario doesn't make much
            #         sense. But maybe stream sometimes returns awaitables and sometimes already
            #         awaited values?
            awaited = in_val  # type: ignore

        self._result_queue.put_nowait(awaited)

    async def _wait_until_empty(self) -> None:
        while True:
            if (
                self._in_stream_exhausted
                and self._result_queue.empty()
                and all(task.done() for task in self._tasks)
            ):
                return
            else:
                await asyncio.sleep(0.01)



from typing import TYPE_CHECKING

from golem.resources.events import NewResource, ResourceClosed, ResourceDataChanged

if TYPE_CHECKING:
    from golem.resources.invoice.invoice import Invoice  # noqa


class NewInvoice(NewResource["Invoice"]):
    pass


class InvoiceDataChanged(ResourceDataChanged["Invoice"]):
    pass


class InvoiceClosed(ResourceClosed["Invoice"]):
    pass



from golem.managers.activity.defaults import default_on_activity_start, default_on_activity_stop
from golem.managers.activity.mixins import ActivityPrepareReleaseMixin
from golem.managers.activity.pool import PoolActivityManager
from golem.managers.activity.single_use import SingleUseActivityManager

__all__ = (
    "default_on_activity_start",
    "default_on_activity_stop",
    "ActivityPrepareReleaseMixin",
    "PoolActivityManager",
    "SingleUseActivityManager",
)



import asyncio
import json
import logging
from datetime import timedelta
from decimal import Decimal
from typing import Dict, List, Optional

from ya_payment import ApiException

from golem.managers.base import ManagerException, PaymentManager
from golem.node import GolemNode
from golem.payload.defaults import DEFAULT_PAYMENT_DRIVER, DEFAULT_PAYMENT_NETWORK
from golem.resources import (
    Agreement,
    Allocation,
    DebitNote,
    Invoice,
    NewAgreement,
    NewDebitNote,
    NewInvoice,
)
from golem.utils.logging import trace_span

logger = logging.getLogger(__name__)

# 2 minutes 30 seconds to as it sometimes take extra time to receive Invoice from an agreement
DEFAULT_SHUTDOWN_TIMEOUT: timedelta = timedelta(seconds=150)


class DefaultPaymentManager(PaymentManager):
    def __init__(
        self,
        golem: GolemNode,
        budget: float,
        network: str = DEFAULT_PAYMENT_NETWORK,
        driver: str = DEFAULT_PAYMENT_DRIVER,
        shutdown_timeout: timedelta = DEFAULT_SHUTDOWN_TIMEOUT,
    ):
        self._golem = golem
        self._budget = budget
        self._network = network
        self._driver = driver
        self._shutdown_timeout = shutdown_timeout.total_seconds()
        self._lock = asyncio.Lock()

        self._allocation: Optional[Allocation] = None

        self._event_handlers: List = []

        self._agreements: Dict[str, Agreement] = {}
        self._no_agreements_event: asyncio.Event = asyncio.Event()
        self._no_agreements_event.set()

    @trace_span("Starting DefaultPaymentManager", log_level=logging.INFO)
    async def start(self):
        self._event_handlers.extend(
            [
                await self._golem.event_bus.on(NewInvoice, self._handle_invoice_payment),
                await self._golem.event_bus.on(NewDebitNote, self._handle_pay_debit_note_payment),
                await self._golem.event_bus.on(NewAgreement, self._handle_new_agreement),
            ]
        )

    @trace_span("Getting allocation", show_results=True, log_level=logging.INFO)
    async def get_allocation(self) -> Allocation:
        async with self._lock:
            if not self._allocation:
                self._allocation = await self._create_allocation(
                    Decimal(self._budget),
                    self._network,
                    self._driver,
                )

        return self._allocation  # type: ignore[return-value]

    async def _release_allocation(self) -> None:
        if self._allocation is None:
            return

        await self._allocation.release()
        self._allocation = None

    @trace_span(show_arguments=True, show_results=True)
    async def _create_allocation(self, budget: Decimal, network: str, driver: str) -> Allocation:
        try:
            return await Allocation.create_any_account(self._golem, budget, network, driver)
        except ApiException as e:
            raise ManagerException(json.loads(e.body)["message"]) from e

    @trace_span("Stopping DefaultPaymentManager", log_level=logging.INFO)
    async def stop(self):
        """Terminate all related agreements."""
        try:
            await asyncio.wait_for(self._wait_for_invoices(), timeout=self._shutdown_timeout)
        except TimeoutError:
            logger.error(
                "Waiting for invoices failed with timeout! Those agreements did not sent invoices:"
                f" {[a for a in self._agreements]}"
            )

        await asyncio.gather(*[agreement.close_all() for agreement in self._agreements.values()])
        self._agreements.clear()

        for event_handler in self._event_handlers:
            await self._golem.event_bus.off(event_handler)
        self._event_handlers.clear()

        await self._release_allocation()

    @trace_span("Waiting for invoices", log_level=logging.INFO)
    async def _wait_for_invoices(self):
        await self._no_agreements_event.wait()

    def _save_agreement(self, agreement: "Agreement") -> None:
        """Add agreement form the pool of known agreements."""
        self._agreements[agreement.id] = agreement
        self._no_agreements_event.clear()
        logger.info(f"Added {agreement.id} to the pool of known agreements")

    def _remove_agreement(self, agreement_id: str) -> None:
        """Remove agreement form the pool of known agreements."""
        del self._agreements[agreement_id]
        if not self._agreements:
            self._no_agreements_event.set()
        logger.info(f"Removed {agreement_id} from the pool of known agreements")

    async def _handle_new_agreement(self, event: NewAgreement) -> None:
        agreement: "Agreement" = event.resource
        self._save_agreement(agreement)

    @trace_span()
    async def _handle_invoice_payment(self, event: NewInvoice) -> None:
        invoice: Invoice = event.resource
        invoice_data = await invoice.get_data()
        if invoice_data.agreement_id not in self._agreements:
            logger.info("Ignoring invoice from unknown agreement `%s`", invoice_data.agreement_id)
            return

        logger.debug("Accepting invoice `%s`: %s", invoice, invoice_data)
        assert self._allocation is not None  # mypy
        await invoice.validate_and_accept(self._allocation)
        self._remove_agreement(invoice_data.agreement_id)

    @trace_span()
    async def _handle_pay_debit_note_payment(self, event: NewDebitNote) -> None:
        debit_note: DebitNote = event.resource
        debit_note_data = await debit_note.get_data()
        if debit_note_data.agreement_id not in self._agreements:
            logger.info(
                "Ignoring debit note from unknown agreement `%s`", debit_note_data.agreement_id
            )
            return

        logger.debug("Accepting debit note `%s`: %s", debit_note, debit_note_data)
        assert self._allocation is not None  # mypy
        await debit_note.validate_and_accept(self._allocation)



from typing import TYPE_CHECKING, Callable, Tuple

from golem.resources.debit_note import DebitNote
from golem.resources.event_collectors import DebitNoteEvent, PaymentEventCollector

if TYPE_CHECKING:
    from golem.resources.activity import Activity


class DebitNoteEventCollector(PaymentEventCollector):
    @property
    def _collect_events_func(self) -> Callable:
        return DebitNote._get_api(self.node).get_debit_note_events

    async def _get_event_resources(self, event: DebitNoteEvent) -> Tuple[DebitNote, "Activity"]:
        assert event.debit_note_id is not None
        debit_note = self.node.debit_note(event.debit_note_id)
        await debit_note.get_data()
        activity = self.node.activity(debit_note.data.activity_id)
        return debit_note, activity



from typing import TYPE_CHECKING

from golem.resources.events import NewResource, ResourceClosed, ResourceDataChanged

if TYPE_CHECKING:
    from golem.resources.network.network import Network  # noqa


class NewNetwork(NewResource["Network"]):
    pass


class NetworkDataChanged(ResourceDataChanged["Network"]):
    pass


class NetworkClosed(ResourceClosed["Network"]):
    pass



import asyncio
import heapq
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import AsyncIterator, Awaitable, Callable, Generic, List, Optional, Tuple, TypeVar

TElement = TypeVar("TElement")


@dataclass(order=True)
class ScoredElement(Generic[TElement]):
    score: float
    element: TElement = field(compare=False)


class Sort:
    """Re-orders elements using a provided scoring function."""

    _no_more_elements: bool = False

    def __init__(
        self,
        score_function: Callable[[TElement], Awaitable[Optional[float]]],
        min_elements: Optional[int] = None,
        max_wait: Optional[timedelta] = None,
        min_wait: Optional[timedelta] = None,
    ):
        """Init Sort.

        :param score_function: Element-scoring function. Higher score -> better element.
            Score `None` indicates an unacceptable element - it will be ignored by the Sort.
        :param min_elements: If not None, :func:`__call__` will not yield anything until
            Sort gathers at least that many elements with a non-None score (but `max_wait` overrides
            this).
        :param max_wait: If not None, we'll not wait for `min_elements` longer than that.
        :param min_wait: If not None, we'll wait for at least that.
        """
        self._score_function = score_function
        self._min_elements = min_elements
        self._max_wait = max_wait
        self._min_wait = min_wait

        self._scored_elements: List[ScoredElement] = []

    async def __call__(self, elements: AsyncIterator[TElement]) -> AsyncIterator[TElement]:
        """Consumes incoming elements as fast as possible. Always yields a element with the \
        highest score.

        :param elements: Stream of element to be reordered.
            In fact, this could be stream of whatever, as long as this whatever matches
            the scoring function, and this whatever would be yielded (TODO - maybe turn this into a
            general Sort?
            https://github.com/golemfactory/golem-core-python/issues/11).
        """
        self._no_more_elements = False
        element_scorer_task = asyncio.get_event_loop().create_task(self._process_stream(elements))
        try:
            element: TElement
            async for element, score in self._elements():
                print(f"Yielding element with score {-1 * score}")
                yield element
        except asyncio.CancelledError:
            element_scorer_task.cancel()
            self._no_more_elements = True

    async def _elements(self) -> AsyncIterator[Tuple[TElement, float]]:
        await self._wait_until_ready()

        while self._scored_elements or not self._no_more_elements:
            try:
                scored_element = heapq.heappop(self._scored_elements)
                yield scored_element.element, scored_element.score
            except IndexError:
                await asyncio.sleep(0.1)

    async def score_function(self, element: TElement) -> Optional[float]:
        return await self._score_function(element)

    async def _process_stream(self, element_stream: AsyncIterator[TElement]) -> None:
        async for element in element_stream:
            #   TODO: https://github.com/golemfactory/golem-core-python/issues/37
            score = await self.score_function(element)
            if score is None:
                continue

            score = score * -1  # heap -> smallest values first -> reverse
            heapq.heappush(self._scored_elements, ScoredElement(score, element))
        self._no_more_elements = True

    async def _wait_until_ready(self) -> None:
        start = datetime.now()

        while True:
            await asyncio.sleep(0.1)
            now = datetime.now()
            if self._min_wait is not None and now - start < self._min_wait:
                # force wait until time exceeds `min_wait`
                continue
            if self._min_elements is None or len(self._scored_elements) >= self._min_elements:
                break
            if self._max_wait is not None and now - start >= self._max_wait:
                break



from golem.resources.agreement.agreement import Agreement
from golem.resources.agreement.events import AgreementClosed, AgreementDataChanged, NewAgreement
from golem.resources.agreement.pipeline import default_create_activity

__all__ = (
    "Agreement",
    "NewAgreement",
    "AgreementDataChanged",
    "AgreementClosed",
    "default_create_activity",
)



from dataclasses import dataclass
from datetime import datetime
from typing import Literal, Optional

from golem.payload import Constraints, Properties

ProposalId = str

# TODO: Use Enum
ProposalState = Literal["Initial", "Draft", "Rejected", "Accepted", "Expired"]


@dataclass
class ProposalData:
    properties: Properties
    constraints: Constraints
    proposal_id: Optional[ProposalId]
    issuer_id: Optional[str]
    state: ProposalState
    timestamp: datetime
    prev_proposal_id: Optional[str]



class GolemException(Exception):
    pass


class MissingConfiguration(GolemException):
    def __init__(self, key: str, description: str):
        self._key = key
        self._description = description

    def __str__(self) -> str:
        return f"Missing configuration for {self._description}. Please set env var {self._key}."



import logging
from typing import Awaitable, Callable, Optional

from golem.event_bus.base import EventBus
from golem.managers.activity.defaults import default_on_activity_start, default_on_activity_stop
from golem.managers.base import WorkContext
from golem.resources import Activity
from golem.utils.logging import trace_span

logger = logging.getLogger(__name__)


class ActivityWrapper:
    def __init__(self, activity: Activity) -> None:
        self._activity = activity

    def __getattr__(self, item):
        return getattr(self._activity, item)


class ActivityPrepareReleaseMixin:
    _event_bus: EventBus

    def __init__(
        self,
        on_activity_start: Optional[
            Callable[[WorkContext], Awaitable[None]]
        ] = default_on_activity_start,
        on_activity_stop: Optional[
            Callable[[WorkContext], Awaitable[None]]
        ] = default_on_activity_stop,
        *args,
        **kwargs,
    ) -> None:
        self._on_activity_start = on_activity_start
        self._on_activity_stop = on_activity_stop

        super().__init__(*args, **kwargs)

    @trace_span(show_arguments=True, show_results=True)
    async def _prepare_activity(self, agreement) -> Activity:
        activity = await agreement.create_activity()
        logger.info(f"Activity `{activity}` created")
        work_context = WorkContext(activity)
        if self._on_activity_start:
            await self._on_activity_start(work_context)
        return activity

    @trace_span(show_arguments=True)
    async def _release_activity(self, activity: Activity) -> None:
        if self._on_activity_stop:
            work_context = WorkContext(activity)
            await self._on_activity_stop(work_context)

        if activity.destroyed:
            logger.info(f"Activity `{activity}` destroyed")
        else:
            logger.warning(
                "ActivityManager expects that activity will be terminated"
                " before activity is released. Looks like you forgot calling"
                " `context.terminate()` in custom `on_activity_end` callback."
            )



from typing import Optional, Sequence, Tuple, Union

from golem.managers.base import ManagerPluginException, ProposalScorer, ProposalScoringResult
from golem.payload.constraints import PropertyName
from golem.resources import ProposalData

PropertyValueNumeric = Union[int, float]
BoundaryValues = Tuple[Tuple[float, PropertyValueNumeric], Tuple[float, PropertyValueNumeric]]


class PropertyValueLerpScore(ProposalScorer):
    """Linear interpolation."""

    def __init__(
        self,
        property_name: PropertyName,
        *,
        minus_one_at: Optional[PropertyValueNumeric] = None,
        zero_at: Optional[PropertyValueNumeric] = None,
        one_at: Optional[PropertyValueNumeric] = None,
        raise_on_missing=False,
        raise_on_bad_value=False,
    ) -> None:
        self._property_name = property_name
        self._boundary_values = self._get_boundary_values(minus_one_at, zero_at, one_at)
        self._raise_on_missing = raise_on_missing
        self._raise_on_bad_value = raise_on_bad_value

    def __call__(self, proposals_data: Sequence[ProposalData]) -> ProposalScoringResult:
        return [self._calculate_linear_score(proposal_data) for proposal_data in proposals_data]

    def _calculate_linear_score(self, proposal_data: ProposalData) -> Optional[float]:
        property_value = self._get_property_value(proposal_data)

        if property_value is None:
            return None

        bounds_min = min([self._boundary_values[0][1], self._boundary_values[1][1]])
        bounds_max = max([self._boundary_values[0][1], self._boundary_values[1][1]])

        x1, y1 = self._boundary_values[0]
        x2, y2 = self._boundary_values[1]
        y3 = min(bounds_max, max(property_value, bounds_min))

        # formula taken from https://www.johndcook.com/interpolatorhelp.html
        return (((y2 - y3) * x1) + ((y3 - y1) * x2)) / (y2 - y1)

    def _get_property_value(self, proposal_data: ProposalData) -> Optional[PropertyValueNumeric]:
        try:
            property_value = proposal_data.properties[self._property_name]
        except KeyError:
            if self._raise_on_missing:
                raise ManagerPluginException(f"Property `{self._property_name}` is not found!")

            return None

        if not isinstance(property_value, (int, float)):
            if self._raise_on_bad_value:
                raise ManagerPluginException(
                    f"Field `{self._property_name}` value type must be an `int` or `float`!"
                )

            return None

        return property_value

    def _get_boundary_values(
        self,
        minus_one_at: Optional[PropertyValueNumeric] = None,
        zero_at: Optional[PropertyValueNumeric] = None,
        one_at: Optional[PropertyValueNumeric] = None,
    ) -> BoundaryValues:
        if minus_one_at is not None:
            bounds_min = (-1, minus_one_at)
        elif zero_at is not None:
            bounds_min = (0, zero_at)
        else:
            raise ManagerPluginException(
                "One of boundary arguments `minus_one_at`, `zero_at` must be provided!"
            )

        if one_at is not None:
            bounds_max = (1, one_at)
        elif zero_at is not None:
            bounds_max = (0, zero_at)
        else:
            raise ManagerPluginException(
                "One of boundary arguments `zero_at`, `one_at` must be provided!"
            )

        return bounds_min, bounds_max



import logging
from typing import Optional

from golem.payload import defaults
from golem.resources import LinearCoeffs, ProposalData
from golem.utils.logging import trace_span

logger = logging.getLogger(__name__)


class LinearCoeffsCost:
    def __init__(self, coeff_name: str) -> None:
        self._coeff_name = coeff_name

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(coeff_name={self._coeff_name})"

    @trace_span(lambda s: str(s), show_results=True)
    def __call__(self, proposal_data: ProposalData) -> Optional[float]:
        coeffs = LinearCoeffs.from_properties(proposal_data.properties)

        if coeffs is None:
            return None

        return float(getattr(coeffs, self._coeff_name))


class LinearPerCpuCoeffsCost(LinearCoeffsCost):
    def __call__(self, proposal_data: ProposalData) -> Optional[float]:
        cpu_count = proposal_data.properties.get(defaults.PROP_INF_CPU_THREADS)

        if not cpu_count:
            return None

        return super().__call__(proposal_data) / cpu_count



from golem.resources.demand.data import DemandData
from golem.resources.demand.demand import Demand
from golem.resources.demand.demand_builder import DemandBuilder
from golem.resources.demand.events import DemandClosed, DemandDataChanged, NewDemand

__all__ = (
    "Demand",
    "DemandBuilder",
    "DemandData",
    "NewDemand",
    "DemandDataChanged",
    "DemandClosed",
)



from golem.managers.proposal.plugins.negotiating.mid_agreement_payments import (
    MidAgreementPaymentsNegotiator,
)
from golem.managers.proposal.plugins.negotiating.negotiating_plugin import NegotiatingPlugin
from golem.managers.proposal.plugins.negotiating.payment_platform import PaymentPlatformNegotiator

__all__ = (
    "PaymentPlatformNegotiator",
    "MidAgreementPaymentsNegotiator",
    "NegotiatingPlugin",
)



import logging
import math
from datetime import timedelta
from typing import Optional

from golem.managers.base import ProposalNegotiator, RejectProposal
from golem.resources import DemandData, ProposalData

logger = logging.getLogger(__name__)

DEFAULT_MIN_DEBIT_NOTE_INTERVAL = timedelta(minutes=1)
DEFAULT_OPTIMAL_DEBIT_NOTE_INTERVAL = timedelta(minutes=10)
DEFAULT_MIN_PAYMENT_TIMEOUT = timedelta(minutes=2)
DEFAULT_OPTIMAL_PAYMENT_TIMEOUT = timedelta(hours=24)

DEFAULT_MIN_ADJUSTMENT = 1
DEFAULT_ADJUSTMENT_FACTOR = 0.33

DEBIT_NOTE_INTERVAL = "golem.com.scheme.payu.debit-note.interval-sec?"
PAYMENT_TIMEOUT = "golem.com.scheme.payu.payment-timeout-sec?"


class MidAgreementPaymentsNegotiator(ProposalNegotiator):
    def __init__(
        self,
        min_debit_note_interval: timedelta = DEFAULT_MIN_DEBIT_NOTE_INTERVAL,
        optimal_debit_note_interval: timedelta = DEFAULT_OPTIMAL_DEBIT_NOTE_INTERVAL,
        min_payment_timeout: timedelta = DEFAULT_MIN_PAYMENT_TIMEOUT,
        optimal_payment_timeout: timedelta = DEFAULT_OPTIMAL_PAYMENT_TIMEOUT,
        min_adjustment: int = DEFAULT_MIN_ADJUSTMENT,
        adjustment_factor: float = DEFAULT_ADJUSTMENT_FACTOR,
    ) -> None:
        self._min_debit_note_interval: int = int(min_debit_note_interval.total_seconds())
        self._optimal_debit_note_interval: int = int(optimal_debit_note_interval.total_seconds())
        self._min_payment_timeout: int = int(min_payment_timeout.total_seconds())
        self._optimal_payment_timeout: int = int(optimal_payment_timeout.total_seconds())
        self._min_adjustment = min_adjustment
        self._adjustment_factor = adjustment_factor

    async def __call__(self, demand_data: DemandData, proposal_data: ProposalData) -> None:
        offer_debit_note_interval = proposal_data.properties.get(DEBIT_NOTE_INTERVAL)
        offer_payment_timeout = proposal_data.properties.get(PAYMENT_TIMEOUT)
        if offer_debit_note_interval is None or offer_payment_timeout is None:
            raise RejectProposal("Offer doesn't support mid agreement payments")

        demand_debit_note_interval = demand_data.properties.get(DEBIT_NOTE_INTERVAL)
        demand_payment_timeout = demand_data.properties.get(PAYMENT_TIMEOUT)

        if (
            offer_debit_note_interval == demand_debit_note_interval
            and offer_payment_timeout == demand_payment_timeout
        ):
            logger.debug(
                "Mid agreement properties negotiation done with debit note interval: "
                f"{offer_debit_note_interval} and payment timeout: {offer_payment_timeout}"
            )
            return

        demand_data.properties[DEBIT_NOTE_INTERVAL] = self._calculate_new_value_proposition(
            offer_debit_note_interval,
            demand_debit_note_interval,
            self._min_debit_note_interval,
            self._optimal_debit_note_interval,
            proposal_data.state == "Initial",
        )
        demand_data.properties[PAYMENT_TIMEOUT] = self._calculate_new_value_proposition(
            offer_payment_timeout,
            demand_payment_timeout,
            self._min_payment_timeout,
            self._optimal_payment_timeout,
            proposal_data.state == "Initial",
        )
        logger.debug(
            "Ongoing mid agreement properties negotiation with"
            f" debit note interval: {demand_data.properties[DEBIT_NOTE_INTERVAL]}"
            f" and payment timeout: {demand_data.properties[PAYMENT_TIMEOUT]}"
        )
        return

    def _calculate_new_value_proposition(
        self,
        offered: int,
        previous: Optional[int],
        minimal: int,
        optimal: int,
        offer_is_initial: bool,
    ):
        # If this is the first proposal,
        # we request the maximum value, unless the offer is already higher.
        if previous is None or offer_is_initial:
            values = [offered, optimal]
            if previous is not None:
                values.append(previous)
            return max(values)
        # If we are offered a higher value, we accept it.
        elif offered >= previous:
            return offered
        # If our proposal was of our minimal value and there is still no consent,
        # we reject the offer.
        elif previous == minimal:
            raise RejectProposal("Offered mid agreement properties are too short")
        # In case of no consent, we adjust the proposed value closer to the offered one.
        else:
            new = previous - max(
                self._min_adjustment, math.ceil((previous - offered) * self._adjustment_factor)
            )
            return max(new, minimal, offered)



import asyncio
import logging
from functools import wraps
from typing import List

from golem.managers.base import (
    WORK_PLUGIN_FIELD_NAME,
    DoWorkCallable,
    Work,
    WorkManagerPlugin,
    WorkResult,
)
from golem.utils.asyncio import ensure_cancelled_many

logger = logging.getLogger(__name__)


def work_plugin(plugin: WorkManagerPlugin):
    def _work_plugin(work: Work):
        if not hasattr(work, WORK_PLUGIN_FIELD_NAME):
            setattr(work, WORK_PLUGIN_FIELD_NAME, [])

        getattr(work, WORK_PLUGIN_FIELD_NAME).append(plugin)

        return work

    return _work_plugin


def retry(tries: int):
    def _retry(do_work: DoWorkCallable) -> DoWorkCallable:
        @wraps(do_work)
        async def wrapper(work: Work) -> WorkResult:
            count = 0
            errors = []
            work_result = WorkResult()

            while count <= tries:
                work_result = await do_work(work)

                if work_result.exception is None:
                    break

                count += 1
                errors.append(work_result.exception)

                logger.info(
                    f"Got an exception {work_result.exception} on {count} attempt {tries-count}"
                    "attempts left"
                )

            work_result.extras["retry"] = {
                "tries": count,
                "errors": errors,
            }

            return work_result

        return wrapper

    return _retry


def redundancy_cancel_others_on_first_done(size: int):
    def _redundancy(do_work: DoWorkCallable):
        @wraps(do_work)
        async def wrapper(work: Work) -> WorkResult:
            tasks: List[asyncio.Task] = [asyncio.ensure_future(do_work(work)) for _ in range(size)]

            tasks_done, tasks_pending = await asyncio.wait(
                tasks, return_when=asyncio.FIRST_COMPLETED
            )

            await ensure_cancelled_many(tasks_pending)

            return tasks_done.pop().result()

        return wrapper

    return _redundancy



import logging
from typing import Set

from golem.managers.base import ProposalNegotiator, RejectProposal
from golem.payload import Properties
from golem.resources import DemandData, ProposalData

logger = logging.getLogger(__name__)


class PaymentPlatformNegotiator(ProposalNegotiator):
    async def __call__(self, demand_data: DemandData, proposal_data: ProposalData) -> None:
        logger.debug("Calling chosen payment platform plugin...")

        if demand_data.properties.get("golem.com.payment.chosen-platform"):
            logger.debug(
                "Calling chosen payment platform plugin done, ignoring as platform already set"
            )
            return

        demand_platforms = self._get_payment_platform_from_properties(demand_data.properties)
        proposal_platforms = self._get_payment_platform_from_properties(proposal_data.properties)
        common_platforms = list(demand_platforms.intersection(proposal_platforms))

        if not common_platforms:
            raise RejectProposal("No common payment platform!")

        chosen_platform = common_platforms[0]

        demand_data.properties["golem.com.payment.chosen-platform"] = chosen_platform

        logger.debug(f"Calling chosen payment platform plugin done with `{chosen_platform}`...")

    def _get_payment_platform_from_properties(self, properties: Properties) -> Set[str]:
        return {
            property.split(".")[4]
            for property in properties
            if property.startswith("golem.com.payment.platform.") and property is not None
        }



import abc
import dataclasses
import enum
from typing import Any, Dict, Final, List, Tuple, Type, TypeVar

from golem.payload.constraints import Constraint, ConstraintOperator, Constraints
from golem.payload.exceptions import InvalidProperties
from golem.payload.properties import Properties

TPayload = TypeVar("TPayload", bound="Payload")

PROP_KEY: Final[str] = "key"
PROP_OPERATOR: Final[str] = "operator"
PROP_MODEL_FIELD_TYPE: Final[str] = "model_field_type"


class PayloadFieldType(enum.Enum):
    constraint = "constraint"
    property = "property"


@dataclasses.dataclass
class Payload(abc.ABC):
    r"""Base class for convenient declaration of Golem's property and constraints syntax.

    Provides helper methods to translate fields between python class and Golem's property and
    constraints syntax.

    Base class for descriptions of the payload required by the requestor.

    example usage::

        import asyncio

        from dataclasses import dataclass
        from golem.resources.market import DemandBuilder, prop, constraint, Payload, RUNTIME_NAME, INF_MEM, INF_STORAGE

        CUSTOM_RUNTIME_NAME = "my-runtime"
        CUSTOM_PROPERTY = "golem.srv.app.myprop"


        @dataclass
        class MyPayload(Payload):
            myprop: str = prop(CUSTOM_PROPERTY, default="myvalue")
            runtime: str = constraint(RUNTIME_NAME, default=CUSTOM_RUNTIME_NAME)
            min_mem_gib: float = constraint(INF_MEM, ">=", default=16)
            min_storage_gib: float = constraint(INF_STORAGE, ">=", default=1024)


        async def main():
            builder = DemandBuilder()
            payload = MyPayload(myprop="othervalue", min_mem_gib=32)
            await builder.add(payload)
            print(builder)

        asyncio.run(main())

    output::

        {'properties': {'golem.srv.app.myprop': 'othervalue'}, 'constraints': ['(&(golem.runtime.name=my-runtime)\n\t(golem.inf.mem.gib>=32)\n\t(golem.inf.storage.gib>=1024))']}



    """  # noqa

    def __init__(self, **kwargs):  # pragma: no cover
        pass

    async def build_properties_and_constraints(self) -> Tuple[Properties, Constraints]:
        return self._build_properties(), self._build_constraints()

    def _build_properties(self) -> Properties:
        """Return a collection of properties declared in the model."""
        return Properties(
            {
                field.metadata[PROP_KEY]: getattr(self, field.name)
                for field in self._get_fields(PayloadFieldType.property)
            }
        )

    def _build_constraints(self) -> Constraints:
        """Return a serialized collection of constraint values."""
        return Constraints(
            [
                Constraint(
                    property_name=field.metadata[PROP_KEY],
                    operator=field.metadata[PROP_OPERATOR],
                    value=getattr(self, field.name),
                )
                for field in self._get_fields(PayloadFieldType.constraint)
            ]
        )

    @classmethod
    def _get_fields(cls, field_type: PayloadFieldType) -> List[dataclasses.Field]:
        """Return a list of fields based on given type."""
        return [
            f
            for f in dataclasses.fields(cls)
            if PROP_KEY in f.metadata and f.metadata.get(PROP_MODEL_FIELD_TYPE) == field_type
        ]

    @classmethod
    def from_properties(cls: Type[TPayload], props: Dict[str, Any]) -> TPayload:
        """Initialize the model with properties from given dictionary.

        Only properties defined in model will be picked up from given dictionary, ignoring other
        keys. In this way several models can be initialized from the same dictionary and all models
        will only load their own data.
        """
        field_map = {
            field.metadata[PROP_KEY]: field for field in cls._get_fields(PayloadFieldType.property)
        }
        data = {
            field_map[key].name: Properties._deserialize_value(val, field_map[key])
            for (key, val) in props.items()
            if key in field_map
        }
        try:
            return cls(**data)
        except TypeError as e:
            raise InvalidProperties(f"Missing key: {e}") from e
        except Exception as e:
            raise InvalidProperties(str(e)) from e


def prop(
    key: str,
    *,
    default: Any = dataclasses.MISSING,
    default_factory: Any = dataclasses.MISSING,
    init: bool = True,
):
    """
    Return a property-type dataclass field for a Payload.

    :param key: the key of the property, e.g. "golem.runtime.name"
    :param default: the default value of the property
    :param default_factory: the default value factory for the property

    example:
    ```python
    >>> from dataclasses import dataclass
    >>> from golem.resources.market import Payload, prop, DemandBuilder
    >>>
    >>> @dataclass
    ... class Foo(Payload):
    ...     bar: int = prop("bar", default=100)
    ...
    >>> builder = DemandBuilder()
    >>> await builder.add(Foo(bar=42))
    >>> print(builder.properties)
    {'bar': 42}
    ```
    """
    return dataclasses.field(  # type: ignore[call-overload]
        default=default,
        default_factory=default_factory,
        metadata={PROP_KEY: key, PROP_MODEL_FIELD_TYPE: PayloadFieldType.property},
        init=init,
    )


def constraint(
    key: str,
    operator: ConstraintOperator = "=",
    *,
    default: Any = dataclasses.MISSING,
    default_factory: Any = dataclasses.MISSING,
    init: bool = True,
):
    """Return a constraint-type dataclass field for a Payload.

    :param key: the key of the property on which the constraint is made - e.g.
        "golem.srv.comp.task_package"
    :param operator: constraint's operator, one of: "=", ">=", "<="
    :param default: the default value for the constraint
    :param default_factory: the default value factory for the constraint

    example:
    ```python
    >>> from dataclasses import dataclass
    >>> from golem.resources.market import Payload, constraint, DemandBuilder
    >>>
    >>> @dataclass
    ... class Foo(Payload):
    ...     max_baz: int = constraint("baz", "<=", default=100)
    ...
    >>> builder = DemandBuilder()
    >>> await builder.add(Foo(max_baz=42))
    >>> print(builder.constraints)
    '(baz<=42)'
    ```
    """
    return dataclasses.field(  # type: ignore[call-overload]
        default=default,
        default_factory=default_factory,
        metadata={
            PROP_KEY: key,
            PROP_OPERATOR: operator,
            PROP_MODEL_FIELD_TYPE: PayloadFieldType.constraint,
        },
        init=init,
    )



import asyncio
import json
from abc import ABC, abstractmethod

# TODO: replace Any here
from typing import Any, Callable, Dict, List, Optional, Union

import aiohttp
import ya_activity
import ya_market
import ya_payment


class YagnaEventCollector(ABC):
    _event_collecting_task: Optional[asyncio.Task] = None

    def start_collecting_events(self) -> None:
        if self._event_collecting_task is None:
            task = asyncio.get_event_loop().create_task(self._collect_yagna_events())
            self._event_collecting_task = task

    def stop_collecting_events(self) -> None:
        if self._event_collecting_task is not None:
            self._event_collecting_task.cancel()
            self._event_collecting_task = None

    async def _collect_yagna_events(self) -> None:
        #   TODO: All of the logic related to "what if collecting fails" is now copied from
        #         yapapi pooling batch logic. Also, is quite ugly.
        #         https://github.com/golemfactory/golem-core-python/issues/50
        gsb_endpoint_not_found_cnt = 0
        MAX_GSB_ENDPOINT_NOT_FOUND_ERRORS = 3

        while True:
            args = self._collect_events_args()
            kwargs = self._collect_events_kwargs()
            try:
                events = await self._collect_events_func(*args, **kwargs)
            except Exception as e:
                if is_intermittent_error(e):
                    continue
                elif is_gsb_endpoint_not_found_error(e):  # type: ignore[arg-type]
                    gsb_endpoint_not_found_cnt += 1
                    if gsb_endpoint_not_found_cnt <= MAX_GSB_ENDPOINT_NOT_FOUND_ERRORS:
                        await asyncio.sleep(3)
                        continue

                raise

            gsb_endpoint_not_found_cnt = 0
            if events:
                for event in events:
                    await self._process_event(event)

    @property
    @abstractmethod
    def _collect_events_func(self) -> Callable:
        raise NotImplementedError

    @abstractmethod
    async def _process_event(self, event: Any) -> None:
        raise NotImplementedError

    def _collect_events_args(self) -> List:
        return []

    def _collect_events_kwargs(self) -> Dict:
        return {}


def is_intermittent_error(e: Exception) -> bool:
    """Check if `e` indicates an intermittent communication failure such as network timeout."""

    is_timeout_exception = isinstance(e, asyncio.TimeoutError) or (
        isinstance(
            e,
            (ya_activity.ApiException, ya_market.ApiException, ya_payment.ApiException),
        )
        and e.status in (408, 504)
    )

    return (
        is_timeout_exception
        or isinstance(e, aiohttp.ServerDisconnectedError)
        # OS error with errno 32 is "Broken pipe"
        or (isinstance(e, aiohttp.ClientOSError) and e.errno == 32)
    )


def is_gsb_endpoint_not_found_error(
    err: Union[ya_activity.ApiException, ya_market.ApiException, ya_payment.ApiException]
) -> bool:
    """Check if `err` is caused by "Endpoint address not found" GSB error."""

    if err.status != 500:
        return False
    try:
        msg = json.loads(err.body)["message"]
        return "GSB error" in msg and "endpoint address not found" in msg
    except Exception:
        return False



from dataclasses import dataclass
from typing import Dict

from golem.payload import defaults


@dataclass
class InfrastructureProps:
    memory_gib: float
    """golem.inf.mem.gib"""
    storage_gib: float
    """golem.inf.storage.gib"""
    cpu_threads: int
    """golem.inf.cpu.threads"""

    @classmethod
    def from_properties(cls, properties: Dict) -> "InfrastructureProps":
        return cls(
            memory_gib=properties.get(defaults.PROP_INF_MEM, 0.0),
            storage_gib=properties.get(defaults.PROP_INF_STORAGE, 0.0),
            cpu_threads=properties.get(defaults.PROP_INF_CPU_THREADS, 0),
        )



import asyncio
import logging
import re
from abc import ABC, ABCMeta
from datetime import datetime, timezone
from functools import wraps
from typing import (
    TYPE_CHECKING,
    AsyncIterator,
    Awaitable,
    Callable,
    Generic,
    Iterable,
    List,
    Optional,
    Type,
    TypeVar,
)

import aiohttp
from typing_extensions import ParamSpec
from ya_activity import ApiException as ActivityApiException
from ya_market import ApiException as MarketApiException
from ya_net import ApiException as NetApiException
from ya_payment import ApiException as PaymentApiException

from golem.resources.events import ResourceDataChanged
from golem.resources.exceptions import ResourceNotFound
from golem.utils.low import TRequestorApi, get_requestor_api

if TYPE_CHECKING:
    from golem.node import GolemNode

logger = logging.getLogger(__name__)

all_api_exceptions = (
    PaymentApiException,
    MarketApiException,
    ActivityApiException,
    NetApiException,
)
P = ParamSpec("P")
R = TypeVar("R")

TResource = TypeVar("TResource", bound="Resource")
TModel = TypeVar("TModel")
TParent = TypeVar("TParent")
TChild = TypeVar("TChild")
TEvent = TypeVar("TEvent")


class _NULL:
    """Set this as a type to tell the typechecker that call is just invalid.

    This might be ugly, but keeps Resource inheritance tree simple.
    """


def api_call_wrapper(
    ignore_status_codes: Iterable[int] = (),
    retry_count: int = 0,
    retry_interval: int = 2,
    retry_status_codes: Iterable[int] = (408, 504),
    retry_exceptions: Iterable[Type[Exception]] = (
        aiohttp.ServerDisconnectedError,
        aiohttp.ClientOSError,
    ),
) -> Callable[[Callable[P, Awaitable[R]]], Callable[P, Awaitable[R]]]:
    def outer_wrapper(f: Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[R]]:
        @wraps(f)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> Optional[R]:
            remaining_tries = retry_count + 1
            while True:
                try:
                    return await f(*args, **kwargs)
                except all_api_exceptions as e:
                    if e.status in ignore_status_codes:
                        return None
                    elif e.status in retry_status_codes or any(
                        isinstance(e, err) for err in retry_exceptions
                    ):
                        remaining_tries -= 1
                        if remaining_tries <= 0:
                            raise
                        await asyncio.sleep(retry_interval)
                    elif e.status == 404:
                        self = args[0]
                        raise ResourceNotFound(self)  # type: ignore
                    else:
                        raise

        return wrapper  # type: ignore  # I don't understand this :/

    return outer_wrapper


class ResourceMeta(ABCMeta):
    """Resources metaclass.

    Ensures a single instance per resource id. Emits the NewResource event.
    """

    def __call__(cls, node: "GolemNode", id_: str, *args, **kwargs):
        assert isinstance(cls, type(Resource))  # mypy
        if args:
            #   Sanity check: when data is passed, it must be a new resource
            assert id_ not in node._resources[cls], f"Repeated id {id_} for class {cls.__name__}"

        if id_ not in node._resources[cls]:
            obj = super(ResourceMeta, cls).__call__(node, id_, *args, **kwargs)  # type: ignore
            node._resources[cls][id_] = obj

            # FIXME: Is there any better solution for this?
            # asyncio.create_task(node.event_bus.emit(NewResource(obj)))

        return node._resources[cls][id_]


class Resource(
    ABC, Generic[TRequestorApi, TModel, TParent, TChild, TEvent], metaclass=ResourceMeta
):
    """Base class of all core-level objects.

    TODO - fix typing in the documentation of this class. I'm afraid this might require
    changes in `sphinx_autodoc_typehints`.

    TODO - in the final version this should disappear from the public docs and all these methods
    should be described on subclasses, but this doesn't make much sense before the previous TODO.
    Related issue: https://github.com/golemfactory/golem-core-python/issues/26
    """

    def __init__(self, node: "GolemNode", id_: str, data: Optional[TModel] = None):
        self._created_at = datetime.now(timezone.utc)
        self._node = node
        self._id = id_
        self._data: Optional[TModel] = data

        self._parent: Optional[TParent] = None
        self._children: List[TChild] = []
        self._events: List[TEvent] = []

        #   When this is done, we know self._children will never change again
        #   This is set by particular resources depending on their internal logic,
        #   and consumed in Resource.child_aiter().
        self._no_more_children: asyncio.Future = asyncio.Future()

        #   Lock for Resource.get_data calls. We don't want to update the same Resource in
        #   multiple tasks at the same time.
        self._get_data_lock = asyncio.Lock()

    ################################
    #   RESOURCE TREE & YAGNA EVENTS
    @property
    def parent(self) -> TParent:
        """Returns a :class:`Resource` we are a child of. Details: :func:`children`."""
        assert self._parent is not None, f"Parent of {self} is not set"
        return self._parent

    @property
    def has_parent(self) -> bool:
        return self._parent is not None

    def add_child(self, child: TChild) -> None:
        assert child._parent is None  # type:ignore
        child._parent = self  # type: ignore
        self._children.append(child)

    @property
    def children(self) -> List[TChild]:
        """List of Resources that were created "from" this resource.

        E.g. children of a :any:`Demand` are :any:`Proposal`.
        Children of :any:`Proposal` are either :any:`Proposal` (counter-proposals to it),
        or :any:`Agreement`.

        """
        return self._children.copy()

    async def child_aiter(self) -> AsyncIterator[TChild]:
        """Yield children. Stops when :class:`Resource` knows there will be no more children."""

        async def no_more_children() -> None:  # missing return statement?
            await self._no_more_children

        stop_task = asyncio.create_task(no_more_children())

        cnt = 0
        while True:
            if cnt < len(self._children):
                yield self._children[cnt]
                cnt += 1
            else:
                #   TODO: https://github.com/golemfactory/golem-core-python/issues/36
                wait_task: asyncio.Task = asyncio.create_task(asyncio.sleep(0.1))
                await asyncio.wait((wait_task, stop_task), return_when=asyncio.FIRST_COMPLETED)
                if stop_task.done():
                    wait_task.cancel()
                    stop_task.result()
                    break

    def add_event(self, event: TEvent) -> None:
        self._events.append(event)

    @property
    def events(self) -> List[TEvent]:
        """Returns a list of all `yagna` events related to this :class:`Resource`.

        Note: these are **yagna** events and should not be confused with `golem.resources.events`.
        """
        return self._events.copy()

    def set_no_more_children(self, exception: Optional[Exception] = None) -> None:
        """Stop :any:`child_aiter` iterator marking as this resource will have no more children.

        This can be called either from iside the resource (e.g. Proposal sets this when it receives
        a ProposalRejected event), or from outside (e.g. on shutdown).
        """
        if not self._no_more_children.done():
            if exception:
                self._no_more_children.set_exception(exception)
            else:
                self._no_more_children.set_result(None)

    ####################
    #   PROPERTIES
    @property
    def api(self) -> TRequestorApi:
        return self._get_api(self.node)

    @property
    def id(self) -> str:
        """Id of the resource, generated by `yagna`."""
        return self._id

    @property
    def data(self) -> TModel:
        """Same as :func:`get_data`, but cached.

        Raises :class:`RuntimeError` if data was never fetched.

        NOTE: `data` might be available even without a prior call to :func:`get_data`
        because some resources (e.g. initial :any:`Proposal`) are fetched from `yagna`
        with full data right away.
        """
        if self._data is None:
            raise RuntimeError(f"Unknown {type(self).__name__} data - call get_data() first")
        return self._data

    @property
    def node(self) -> "GolemNode":
        """:any:`GolemNode` that defines the context of this :class:`Resource`."""
        return self._node

    @property
    def created_at(self) -> datetime:
        return self._created_at

    ####################
    #   DATA LOADING
    async def get_data(self, force: bool = False) -> TModel:
        """Return details of this resource.

        :param force: False -> returns the cached data (or fetches data from `yagna` if there is no
            cached version). True -> always fetches the new data (and updates the cache).

        """
        async with self._get_data_lock:
            if self._data is None or force:
                old_data = self._data
                self._data = await self._get_data()
                if old_data is not None and old_data != self._data:
                    await self.node.event_bus.emit(ResourceDataChanged(self, old_data))

        assert self._data is not None  # mypy
        return self._data

    @api_call_wrapper()
    async def _get_data(self) -> TModel:
        get_method: Callable[[str], Awaitable[TModel]] = getattr(self.api, self._get_method_name)
        return await get_method(self._id)

    @classmethod
    @api_call_wrapper()
    async def get_all(cls: Type[TResource], node: "GolemNode") -> List[TResource]:
        api = cls._get_api(node)
        get_all_method = getattr(api, cls._get_all_method_name())
        data = await get_all_method()

        resources = []
        id_field = cls._id_field_name()
        for raw in data:
            id_ = getattr(raw, id_field)
            resources.append(cls(node, id_, raw))
        return resources

    ###################
    #   OTHER
    @classmethod
    def _get_api(cls, node: "GolemNode") -> TRequestorApi:
        return get_requestor_api(cls, node)

    @property
    def _get_method_name(self) -> str:
        """Name of the single GET ya_client method, e.g. get_allocation."""
        return f"get_{self._snake_case_name()}"

    @classmethod
    def _get_all_method_name(cls) -> str:
        """Name of the collection GET ya_client method, e.g. get_allocations."""
        return f"get_{cls._snake_case_name()}s"

    @classmethod
    def _id_field_name(cls) -> str:
        return f"{cls._snake_case_name()}_id"

    @classmethod
    def _snake_case_name(cls) -> str:
        replaced = re.sub("([A-Z]+)", r"_\1", cls.__name__).lower()
        return replaced[1:]

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._id})"



from golem.resources.activity.activity import Activity
from golem.resources.activity.commands import (
    Command,
    Deploy,
    DownloadFile,
    Run,
    Script,
    SendFile,
    Start,
)
from golem.resources.activity.events import ActivityClosed, ActivityDataChanged, NewActivity
from golem.resources.activity.pipeline import default_prepare_activity

__all__ = (
    "Activity",
    "NewActivity",
    "ActivityDataChanged",
    "ActivityClosed",
    "Command",
    "Script",
    "Deploy",
    "Start",
    "Run",
    "SendFile",
    "DownloadFile",
    "default_prepare_activity",
)



import asyncio
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, List, Optional

from ya_market import RequestorApi
from ya_market import models as models
from ya_market.exceptions import ApiException

from golem.resources.activity import Activity
from golem.resources.agreement.data import AgreementData
from golem.resources.agreement.events import AgreementClosed, NewAgreement
from golem.resources.base import _NULL, Resource, api_call_wrapper
from golem.resources.invoice import Invoice

if TYPE_CHECKING:
    from golem.node import GolemNode
    from golem.resources.proposal import Proposal  # noqa


class Agreement(Resource[RequestorApi, models.Agreement, "Proposal", Activity, _NULL]):
    """A single agreement on the Golem Network.

    Sample usage::

        agreement = await proposal.create_agreement()
        await agreement.confirm()
        await agreement.wait_for_approval()
        activity = await agreement.create_activity()
        # Use the activity
        await agreement.terminate()
    """

    def __init__(self, node: "GolemNode", id_: str, data: Optional[models.Agreement] = None):
        super().__init__(node, id_, data)
        self._approved_at: Optional[datetime] = None
        asyncio.create_task(node.event_bus.emit(NewAgreement(self)))

    @api_call_wrapper()
    async def confirm(self) -> None:
        """Confirm the agreement.

        First step that leads to an active agreement.
        """
        await self.api.confirm_agreement(self.id, app_session_id=self.node.app_session_id)

    @api_call_wrapper()
    async def wait_for_approval(self) -> bool:
        """Wait for provider's approval of the agreement.

        Second (and last) step leading to an active agreement.

        :returns: True if agreement was approved.
        """
        try:
            approved_at = datetime.now(timezone.utc)
            await self.api.wait_for_approval(self.id, timeout=15, _request_timeout=16)
            self._approved_at = approved_at
            return True
        except ApiException as e:
            if e.status == 410:
                return False
            elif e.status == 408:
                #   TODO: maybe this should be in api_call_wrapper?
                return await self.wait_for_approval()
            else:
                raise

    @api_call_wrapper()
    async def create_activity(
        self, autoclose: bool = True, timeout: timedelta = timedelta(seconds=10)
    ) -> "Activity":
        """Create a new :any:`Activity` for this :any:`Agreement`.

        :param autoclose: Destroy the activity when the :any:`GolemNode` closes.
        :param timeout: Request timeout.
        """
        from golem.resources.activity.activity import Activity

        activity = await Activity.create(self.node, self.id, timeout)
        if autoclose:
            self.node.add_autoclose_resource(activity)
        self.add_child(activity)
        return activity

    @api_call_wrapper()
    async def terminate(self, reason: str = "") -> None:
        """Terminate the agreement.

        :param reason: Optional information for the provider explaining why the agreement was
            terminated.
        """
        try:
            await self.api.terminate_agreement(self.id, request_body={"message": reason})
            await self.node.event_bus.emit(AgreementClosed(self))
        except ApiException as e:
            if self._is_permanent_410(e):
                pass
            else:
                raise

    @property
    def approved_at(self) -> Optional[datetime]:
        return self._approved_at

    @property
    def invoice(self) -> Optional[Invoice]:
        """:any:`Invoice` for this :any:`Agreement`, or None if we didn't yet receive an invoice."""
        try:
            return [child for child in self.children if isinstance(child, Invoice)][0]
        except IndexError:
            return None

    @property
    def activities(self) -> List["Activity"]:
        """A list of :any:`Activity` created for this :any:`Agreement`."""
        return [child for child in self.children if isinstance(child, Activity)]

    async def close_all(self) -> None:
        """Terminate agreement, destroy all activities.

        Ensure success -> retry if there are any problems.

        This is indended to be used in scenarios when we just want to end
        this agreement and we want to make sure it is really terminated (even if e.g. in some other
        separate task we're waiting for the provider to approve it).
        """
        #   TODO: This method is very ugly, also similar method could be useful for acivity only.
        #   BUT this probably should be a yagna-side change. Agreement.terminate() should
        #   just always succeed, as well as Activity.destroy() - yagna should repeat if necessary
        #   etc. We should only repeat in rare cases when we can't connect to our local `yagna`.
        #   Related issue: https://github.com/golemfactory/golem-core-python/issues/19

        #   Q: Why limit on repeats?
        #   A: So that we don't flood `yagna` with requests that will never succeed.
        #   Q: Why repeating 4 times?
        #   A: No particular reason.

        for i in range(1, 5):
            try:
                await self.terminate()
                break
            except ApiException as e:
                if self._is_permanent_410(e):
                    break
            await asyncio.sleep(2**i)

        for activity in self.activities:
            for i in range(1, 5):
                try:
                    await activity.destroy()
                    break
                except Exception:
                    pass
                await asyncio.sleep(2**i)

    @staticmethod
    def _is_permanent_410(e: ApiException) -> bool:
        #   TODO: Remove this check once https://github.com/golemfactory/yagna/issues/2264 is done
        #         and every 410 is permanent.
        if e.status != 410:
            return False
        return "from Approving" not in str(e) and "from Pending" not in str(e)

    @property
    def proposal(self) -> "Proposal":
        return self.parent

    async def get_agreement_data(self, force=False) -> AgreementData:
        data = await self.get_data(force=force)
        agreement_duration = (
            datetime.now(timezone.utc) - self.approved_at if self.approved_at else None
        )
        return AgreementData(
            agreement_id=data.agreement_id,
            provider_id=data.offer.provider_id,
            approved_date=data.approved_date,
            properties=data.offer.properties,  # type: ignore
            agreement_duration=agreement_duration,
        )



from golem.managers.demand.aggregating import AggregatingDemandManager
from golem.managers.demand.refreshing import RefreshingDemandManager

__all__ = (
    "AggregatingDemandManager",
    "RefreshingDemandManager",
)



import logging
from abc import ABC
from dataclasses import dataclass
from datetime import timedelta
from enum import Enum
from typing import Final, List, Literal, Optional, Tuple

from dns.exception import DNSException
from srvresolver.srv_record import SRVRecord
from srvresolver.srv_resolver import SRVResolver

from golem.payload import defaults
from golem.payload.base import Payload, constraint, prop
from golem.payload.constraints import Constraints
from golem.payload.properties import Properties
from golem.utils.http import make_http_get_request, make_http_head_request

DEFAULT_REPO_URL_SRV: Final[str] = "_girepo._tcp.dev.golem.network"
DEFAULT_REPO_URL_FALLBACK: Final[str] = "http://girepo.dev.golem.network:8000"
DEFAULT_REPO_URL_TIMEOUT: Final[timedelta] = timedelta(seconds=10)

VmCaps = Literal["vpn", "inet", "manifest-support"]

logger = logging.getLogger(__name__)


class VmPayloadException(Exception):
    """Exception raised on any problems related to the payload for "vm" runtime."""


class VmPackageFormat(Enum):
    """Enumeration of available image formats for "vm" runtime."""

    UNKNOWN = None
    GVMKIT_SQUASH = "gvmkit-squash"


@dataclass
class BaseVmPayload(Payload, ABC):
    """Declarative description of common payload parameters for "vm" runtime."""

    runtime: str = constraint(defaults.PROP_RUNTIME_NAME, "=", default="vm")
    capabilities: List[VmCaps] = constraint(
        defaults.PROP_RUNTIME_CAPABILITIES, "=", default_factory=list
    )

    min_mem_gib: float = constraint(defaults.PROP_INF_MEM, ">=", default=0.5)
    min_storage_gib: float = constraint(defaults.PROP_INF_STORAGE, ">=", default=2.0)
    min_cpu_threads: int = constraint(defaults.PROP_INF_CPU_THREADS, ">=", default=1)

    package_format: VmPackageFormat = prop(
        "golem.srv.comp.vm.package_format", default=VmPackageFormat.GVMKIT_SQUASH
    )


@dataclass
class _VmPayload(Payload, ABC):
    package_url: str = prop("golem.srv.comp.task_package")


# TODO: Use kw_only=True on python 3.10+
@dataclass
class VmPayload(BaseVmPayload, _VmPayload):
    """Declarative description of payload for "vm" runtime."""


@dataclass
class _ManifestVmPayload(Payload, ABC):
    manifest: str = prop("golem.srv.comp.payload")
    manifest_sig: Optional[str] = prop("golem.srv.comp.payload.sig", default=None)
    manifest_sig_algorithm: Optional[str] = prop(
        "golem.srv.comp.payload.sig.algorithm", default=None
    )
    manifest_cert: Optional[str] = prop("golem.srv.comp.payload.cert", default=None)


# TODO: Use kw_only=True on python 3.10+
@dataclass
class ManifestVmPayload(BaseVmPayload, _ManifestVmPayload):
    """Declarative description of payload for "vm" runtime, with Computation Manifest flavor."""


@dataclass
class _RepositoryVmPayload(ABC):
    image_hash: str
    image_url: Optional[str] = None
    package_url: Optional[str] = prop("golem.srv.comp.task_package", default=None)


# TODO: Use kw_only=True on python 3.10+
@dataclass
class RepositoryVmPayload(BaseVmPayload, _RepositoryVmPayload):
    """Declarative description of payload for "vm" runtime, with ability to resolve "image_url" \
    parameter from remote repository."""

    async def _resolve_package_url(self) -> None:
        if self.image_url:
            await check_image_url(self.image_url)
            image_url = self.image_url
        else:
            repo_url = await resolve_repository_url()
            image_url = await resolve_image_url(repo_url, self.image_hash)

        self.package_url = get_package_url(self.image_hash, image_url)

    async def build_properties_and_constraints(self) -> Tuple[Properties, Constraints]:
        if self.package_url is None:
            await self._resolve_package_url()

        return await super().build_properties_and_constraints()


async def resolve_repository_url(
    repo_srv: str = DEFAULT_REPO_URL_SRV,
    fallback_url: str = DEFAULT_REPO_URL_FALLBACK,
    timeout: timedelta = DEFAULT_REPO_URL_TIMEOUT,
) -> str:
    """Resolve url of the image repository with help of DNS records.

    :param repo_srv: the SRV record that keep repository url
    :param fallback_url: fallback url in case there's a problem resolving SRV record
    :param timeout: maximum time window for SRV resolving
    :return: the full repository url
    """
    try:
        try:
            # TODO: Async implementation is needed
            srv: Optional[SRVRecord] = SRVResolver.resolve_random(
                repo_srv,
                timeout=int(timeout.total_seconds()),
            )
        except DNSException as e:
            raise VmPayloadException("Could not resolve Golem package repository address!") from e

        if not srv:
            raise VmPayloadException("Golem package repository is currently unavailable!")
    except Exception as e:
        # this is a temporary fallback for a problem resolving the SRV record
        logger.warning(
            "Problem resolving %s, falling back to %s, exception: %s", repo_srv, fallback_url, e
        )
        return fallback_url

    return f"http://{srv.host}:{srv.port}"


async def check_image_url(image_url: str) -> None:
    """Check if given image url exists."""

    await make_http_head_request(image_url, raise_exceptions=True)


async def resolve_image_url(repo_url: str, image_hash: str) -> str:
    """Fetch image url from given repository url and image hash."""

    return await make_http_get_request(f"{repo_url}/image.{image_hash}.link", raise_exceptions=True)


def get_package_url(image_hash: str, image_url: str) -> str:
    """Turn image hash and image url into package url format."""
    return f"hash:sha3:{image_hash}:{image_url}"



from golem.managers.proposal.default import DefaultProposalManager
from golem.managers.proposal.plugins import (
    BlacklistProviderIdPlugin,
    LinearAverageCostPricing,
    LinearCoeffsCost,
    LinearPerCpuAverageCostPricing,
    LinearPerCpuCoeffsCost,
    MapScore,
    MidAgreementPaymentsNegotiator,
    NegotiatingPlugin,
    PaymentPlatformNegotiator,
    PropertyValueLerpScore,
    ProposalBuffer,
    ProposalScoringBuffer,
    ProposalScoringMixin,
    RandomScore,
    RejectIfCostsExceeds,
)

__all__ = (
    "DefaultProposalManager",
    "BlacklistProviderIdPlugin",
    "ProposalBuffer",
    "PaymentPlatformNegotiator",
    "MidAgreementPaymentsNegotiator",
    "NegotiatingPlugin",
    "RejectIfCostsExceeds",
    "MapScore",
    "ProposalScoringMixin",
    "LinearAverageCostPricing",
    "LinearPerCpuAverageCostPricing",
    "LinearCoeffsCost",
    "LinearPerCpuCoeffsCost",
    "PropertyValueLerpScore",
    "RandomScore",
    "ProposalScoringBuffer",
)



import asyncio
import json
import logging
from datetime import timedelta
from typing import TYPE_CHECKING, List, Optional

from ya_activity import models

from golem.resources.activity.commands import Command, Script
from golem.resources.activity.events import ActivityClosed, NewActivity
from golem.resources.base import _NULL, Resource, api_call_wrapper
from golem.resources.debit_note import DebitNote
from golem.resources.pooling_batch import PoolingBatch
from golem.utils.low import ActivityApi

if TYPE_CHECKING:
    from golem.node import GolemNode
    from golem.resources.agreement import Agreement  # noqa

logger = logging.getLogger(__name__)


class Activity(Resource[ActivityApi, _NULL, "Agreement", PoolingBatch, _NULL]):
    """A single activity on the Golem Network.

    Either created by :any:`Agreement.create_activity()` or via :any:`GolemNode.activity()`.
    """

    def __init__(self, node: "GolemNode", id_: str):
        super().__init__(node, id_)

        asyncio.create_task(node.event_bus.emit(NewActivity(self)))

        self._running_batch_counter = 0
        self._busy_event = asyncio.Event()
        self._idle_event = asyncio.Event()
        self._idle_event.set()
        self._destroyed_event = asyncio.Event()

    ###################
    #   State management - idle / busy / destroyed
    @property
    def idle(self) -> bool:
        """True if there are no batches being executed now on this :any:`Activity`."""
        return self._idle_event.is_set()

    @property
    def running_batch_counter(self) -> int:
        return self._running_batch_counter

    @running_batch_counter.setter
    def running_batch_counter(self, new_val: int) -> None:
        assert abs(self._running_batch_counter - new_val) == 1  # change by max 1
        assert new_val >= 0
        self._running_batch_counter = new_val
        if new_val == 0:
            self._busy_event.clear()
            self._idle_event.set()
        else:
            self._busy_event.set()
            self._idle_event.clear()

    async def wait_busy(self) -> None:
        """Wait until this :any:`Activity` is no longer :any:`idle`."""
        await self._busy_event.wait()

    async def wait_idle(self) -> None:
        """Wait until this :any:`Activity` is :any:`idle`."""
        await self._idle_event.wait()

    async def wait_destroyed(self) -> None:
        """Wait until this :any:`Activity` is :any:`destroyed`."""
        await self._destroyed_event.wait()

    @property
    def destroyed(self) -> bool:
        """True after a successful call to :func:`destroy`."""
        return self._destroyed_event.is_set()

    ####################
    #   API
    @api_call_wrapper(retry_count=2)
    async def get_state(self) -> models.ActivityState:
        """Get state of the Activity."""
        return await self.api.get_activity_state(self.id)

    @classmethod
    async def create(cls, node: "GolemNode", agreement_id: str, timeout: timedelta) -> "Activity":
        api = cls._get_api(node)
        activity_id = await api.create_activity(agreement_id, timeout=timeout.total_seconds())
        activity = cls(node, activity_id)
        return activity

    @api_call_wrapper()
    async def destroy(self) -> None:
        """Destroy this :any:`Activity`. This is final, destroyed activities can no longer be \
        used."""
        await self.api.destroy_activity(self.id)
        self._destroyed_event.set()
        await self.node.event_bus.emit(ActivityClosed(self))

    @api_call_wrapper()
    async def execute(
        self, script: models.ExeScriptRequest, timeout: Optional[float] = None
    ) -> PoolingBatch:
        batch_id = await self.api.call_exec(self.id, script, _request_timeout=timeout)
        batch = PoolingBatch(self.node, batch_id)
        batch.start_collecting_events()
        self.add_child(batch)
        self.running_batch_counter += 1
        return batch

    async def execute_commands(
        self, *commands: Command, timeout: Optional[float] = None
    ) -> PoolingBatch:
        """Create a new batch that executes given :any:`Command` s in the exe unit.

        Sample usage::

            batch = await activity.execute_commands(
                Deploy(),
                Start(),
                Run("echo -n 'hello world'"),
            )
            await batch.wait()
            print(batch.events[-1].stdout)  # "hello world"
        """
        await asyncio.gather(*[c.before() for c in commands])
        commands_str = json.dumps([c.text() for c in commands])
        batch = await self.execute(models.ExeScriptRequest(text=commands_str), timeout)

        async def execute_after() -> None:
            await batch.wait(ignore_errors=True)
            await asyncio.gather(*[c.after() for c in commands])

        asyncio.create_task(execute_after())
        return batch

    async def execute_script(self, script: "Script") -> PoolingBatch:
        """Create a new batch that executes commands from a given :any:`Script` in the exe unit.

        This is an alternative to :func:`execute_commands` that provides a more granular access to
        the results.

        Sample usage::

            script = Script()
            script.add_command(Deploy())
            script.add_command(Start())
            result = script.add_command(Run("echo -n 'hello world'"))
            script.add_command(Run("sleep 1000"))

            batch = await activity.execute_script(script)

            #   This line doesn't wait for the whole batch to finish
            print((await result).stdout)  # "hello world"
        """
        batch = await self.execute_commands(*script.commands)
        batch._futures = script.futures
        return batch

    def batch(self, batch_id: str) -> PoolingBatch:
        """Return a :any:`PoolingBatch` with a given id.

        Id is assume to be correct, there is no validation.
        """
        batch = PoolingBatch(self.node, batch_id)
        if batch._parent is None:
            self.add_child(batch)
        return batch

    @property
    def debit_notes(self) -> List[DebitNote]:
        """List of all debit notes for this :any:`Activity`."""
        return [child for child in self.children if isinstance(child, DebitNote)]

    @property
    def agreement(self) -> "Agreement":
        return self.parent



from copy import deepcopy
from typing import TYPE_CHECKING, Optional, Union

from golem.payload import Payload
from golem.payload.constraints import Constraint, ConstraintGroup, Constraints
from golem.payload.properties import Properties
from golem.resources.demand.demand import Demand

if TYPE_CHECKING:
    from golem.node import GolemNode


class DemandBuilder:
    """Builder that gives ability to collect Demand's properties and constraints in multiple \
    separate steps, prior its creation.

    example usage:

    ```python
    >>> from golem.resources import DemandBuilder
    >>> from golem.payload import defaults
    >>> from datetime import datetime, timezone
    >>> builder = DemandBuilder()
    >>> await builder.add(defaults.NodeInfo(name="a node", subnet_tag="testnet"))
    >>> await builder.add(defaults.ActivityInfo(expiration=datetime.now(timezone.utc)))
    >>> print(builder)
    {'properties':
        {'golem.node.id.name': 'a node',
         'golem.node.debug.subnet': 'testnet',
         'golem.srv.comp.expiration': 1601655628772},
     'constraints': []}
    ```
    """

    def __init__(
        self, properties: Optional[Properties] = None, constraints: Optional[Constraints] = None
    ):
        self.properties: Properties = (
            deepcopy(properties) if properties is not None else Properties()
        )
        self.constraints: Constraints = (
            deepcopy(constraints) if constraints is not None else Constraints()
        )

    def __repr__(self):
        return repr({"properties": self.properties, "constraints": self.constraints})

    def __eq__(self, other):
        return (
            isinstance(other, DemandBuilder)
            and self.properties == other.properties
            and self.constraints == other.constraints
        )

    async def add(self, payload: Payload):
        """Add properties and constraints from the given payload to this demand definition."""

        properties, constraints = await payload.build_properties_and_constraints()

        self.add_properties(properties)
        self.add_constraints(constraints)

    def add_properties(self, props: Properties):
        """Add properties from the given dictionary to this demand definition."""
        self.properties.update(props)

    def add_constraints(self, constraints: Union[Constraint, ConstraintGroup]):
        """Add a constraint from given args to the demand definition."""
        if (
            isinstance(constraints, ConstraintGroup)
            and constraints.operator == self.constraints.operator
        ):
            self.constraints.items.extend(constraints.items)
        else:
            self.constraints.items.append(constraints)

    async def create_demand(self, node: "GolemNode") -> "Demand":
        """Create demand and subscribe to its events."""
        return await Demand.create_from_properties_constraints(
            node, self.properties, self.constraints
        )



from golem.resources.allocation.allocation import Allocation
from golem.resources.allocation.events import AllocationClosed, AllocationDataChanged, NewAllocation
from golem.resources.allocation.exceptions import AllocationException, NoMatchingAccount

__all__ = (
    "Allocation",
    "AllocationException",
    "NoMatchingAccount",
    "NewAllocation",
    "AllocationDataChanged",
    "AllocationClosed",
)



import logging
from dataclasses import dataclass
from datetime import timedelta
from decimal import ROUND_CEILING, Decimal
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Union

from golem.payload import defaults
from golem.resources.exceptions import PaymentValidationException
from golem.resources.utils.infrastructure import InfrastructureProps

logger = logging.getLogger(__name__)


ETH_EXPONENT = 10 ** Decimal(-18)


USAGE_VECTOR_TO_PRICE_MAPPING = {
    # `_` versions are deprecated but still used by providers
    "golem.usage.cpu_sec": "price_cpu_sec",
    "golem.usage.cpu-sec": "price_cpu_sec",
    "golem.usage.duration_sec": "price_duration_sec",
    "golem.usage.duration-sec": "price_duration_sec",
    "golem.usage.storage_gib": "price_storage_gib",
    "golem.usage.storage-gib": "price_storage_gib",
    "golem.usage.gib": "price_mem_gib",
}


@dataclass
class LinearCoeffs:
    usage_vector: List[str]
    price_storage_gib: Decimal = Decimal("0.0")
    price_mem_gib: Decimal = Decimal("0.0")
    price_cpu_sec: Decimal = Decimal("0.0")
    price_duration_sec: Decimal = Decimal("0.0")
    price_initial: Decimal = Decimal("0.0")

    def usage_vector_price(self) -> List[Decimal]:
        return [
            getattr(self, USAGE_VECTOR_TO_PRICE_MAPPING[vector])
            for vector in self.usage_vector
            if vector in USAGE_VECTOR_TO_PRICE_MAPPING
        ]

    @classmethod
    def from_properties(cls, properties: Dict) -> Optional["LinearCoeffs"]:
        pricing_model = properties.get(defaults.PROP_PRICING_MODEL)

        if pricing_model != "linear":
            logger.debug("Pricing model `%s` is not `linear`, ignoring", pricing_model)
            return None

        coeffs: List[Decimal] = [
            eth_decimal(coeff) for coeff in properties.get(defaults.PROP_PRICING_LINEAR_COEFFS, [])
        ]
        usage_vector: List[str] = [
            usage for usage in properties.get(defaults.PROP_USAGE_VECTOR, [])
        ]

        if len(coeffs) == 0 or len(coeffs) != len(usage_vector) + 1:
            logger.debug(
                "Linear pricing coeffs must be a sequence matching a length of usage vector + 1"
                "`%s` != `%s`, ignoring",
                len(coeffs),
                len(usage_vector) + 1,
            )

            return None
        build_dict: Dict[str, Any] = {"price_initial": coeffs[-1], "usage_vector": usage_vector}

        for usage, coeff in zip(usage_vector, coeffs):
            key = USAGE_VECTOR_TO_PRICE_MAPPING.get(usage)
            if key is None:
                continue
            build_dict[key] = coeff

        return cls(**build_dict)


class PayDocumentStatus(str, Enum):
    ISSUED = "ISSUED"
    RECEIVED = "RECEIVED"
    ACCEPTED = "ACCEPTED"
    REJECTED = "REJECTED"
    FAILED = "FAILED"
    SETTLED = "SETTLED"
    CANCELLED = "CANCELLED"


@dataclass
class PaymentProps:
    debit_notes_accept_timeout: Optional[int] = None
    """golem.com.payment.debit-notes.accept-timeout?"""
    debit_note_interval: Optional[int] = None
    """golem.com.scheme.payu.debit-note.interval-sec?"""
    payment_timeout: Optional[int] = None
    """golem.com.scheme.payu.payment-timeout-sec?"""

    @classmethod
    def from_properties(cls, properties: Dict) -> "PaymentProps":
        return cls(
            debit_notes_accept_timeout=properties.get(defaults.PROP_DEBIT_NOTES_ACCEPT_TIMEOUT),
            debit_note_interval=properties.get(defaults.PROP_DEBIT_NOTES_INTERVAL),
            payment_timeout=properties.get(defaults.PROP_PAYMENT_TIMEOUT),
        )


def eth_decimal(value: Union[str, Decimal]) -> Decimal:
    return Decimal(value).quantize(ETH_EXPONENT, ROUND_CEILING)


def validate_payment_max_cost(
    coeffs: LinearCoeffs,
    inf: InfrastructureProps,
    duration: timedelta,
    amount: Decimal,
    time_since_last_debit_note: Optional[timedelta] = None,
    amount_since_last_debit_note: Optional[Decimal] = None,
    grace_period: timedelta = timedelta(minutes=1),
) -> Tuple[Decimal, Optional[Decimal]]:
    """Validate payment data max cost.

    Returns:
        maximum cost given `coeffs` and `infrastructure` could generate in
        - given `duration`
        - given `time_since_last_debit_note` if provided

    Raises: PaymentValidationException
    """
    duration_grace = duration + grace_period
    max_cost = eth_decimal(
        coeffs.price_storage_gib * Decimal(inf.storage_gib)
        + coeffs.price_mem_gib * Decimal(inf.memory_gib)
        + coeffs.price_cpu_sec * Decimal(inf.cpu_threads) * Decimal(duration_grace.total_seconds())
        + coeffs.price_duration_sec * Decimal(duration_grace.total_seconds())
        + coeffs.price_initial
    )

    if amount > max_cost:
        raise PaymentValidationException(
            "Total amount due exceeds expected max possible cost " f"{amount} > {max_cost}"
        )

    if time_since_last_debit_note is None or amount_since_last_debit_note is None:
        return max_cost, None

    max_cost_since_last_debit_note = eth_decimal(
        coeffs.price_storage_gib * Decimal(inf.storage_gib)
        + coeffs.price_mem_gib * Decimal(inf.memory_gib)
        + coeffs.price_cpu_sec
        * Decimal(inf.cpu_threads)
        * Decimal(time_since_last_debit_note.total_seconds())
        + coeffs.price_duration_sec * Decimal(time_since_last_debit_note.total_seconds())
        + coeffs.price_initial
    )

    if amount_since_last_debit_note > max_cost_since_last_debit_note:
        raise PaymentValidationException(
            "Amount due since last debit note exceeds expected max possible cost "
            f"{amount_since_last_debit_note} > {max_cost_since_last_debit_note}"
        )

    return max_cost, max_cost_since_last_debit_note


def validate_payment_calculated_cost(
    coeffs: LinearCoeffs,
    amount: Decimal,
    usage_counter_vector: List,
    grace_amount: Decimal = 100 * ETH_EXPONENT,
) -> Decimal:
    """Validate payment amount calculated from vector usage.

    Returns: cost given `coeffs` and `usage_counter_vector` should generate

    Raises: PaymentValidationException
    """
    calculated_cost = coeffs.price_initial
    for price, value in zip(coeffs.usage_vector_price(), usage_counter_vector):
        calculated_cost += price * Decimal(value)
    calculated_cost = eth_decimal(calculated_cost)

    if amount > calculated_cost + grace_amount:
        raise PaymentValidationException(
            "Total amount due exceeds expected calculated cost " f"{amount} > {calculated_cost}"
        )

    return calculated_cost



import asyncio
from typing import Generic, Optional, TypeVar

from golem.utils.asyncio.tasks import ensure_cancelled_many

TQueueItem = TypeVar("TQueueItem")


class ErrorReportingQueue(asyncio.Queue, Generic[TQueueItem]):
    """Asyncio Queue that enables exceptions to be passed to consumers from the feeding code."""

    def __init__(self, *args, **kwargs):
        self._error: Optional[BaseException] = None
        self._error_event: asyncio.Event = asyncio.Event()

        super().__init__(*args, **kwargs)

    def get_nowait(self) -> TQueueItem:
        """Perform a regular `get_nowait` if there are items in the queue.

        Otherwise, if an exception had been signalled, raise the exception.
        """
        if self.empty() and self._error:
            raise self._error
        return super().get_nowait()

    async def get(self) -> TQueueItem:
        """Perform a regular, waiting `get` but raise an exception if happens while waiting.

        If there had been items in the queue,
        they will first be returned before an exception is raised.
        """

        error_task = asyncio.create_task(self._error_event.wait())
        get_task = asyncio.create_task(super().get())
        done, pending = await asyncio.wait(
            [error_task, get_task], return_when=asyncio.FIRST_COMPLETED
        )

        await ensure_cancelled_many(pending)

        if get_task in done:
            return await get_task

        assert self._error
        raise self._error

    async def put(self, item: TQueueItem):
        await super().put(item)

    def put_nowait(self, item: TQueueItem):
        super().put_nowait(item)

    def set_exception(self, exc: BaseException):
        """Set the exception, causing subsequent get calls to fail with an exception."""
        self._error = exc
        self._error_event.set()

    def reset_exception(self):
        """Reset the exception in case the condition that caused it originally had been resolved."""
        self._error = None
        self._error_event.clear()



from abc import ABC, abstractmethod
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any, Dict, Tuple, Union

from ya_payment import models

from golem.resources.base import Resource
from golem.utils.low import YagnaEventCollector

if TYPE_CHECKING:
    from golem.node import GolemNode

InvoiceEvent = Union[
    models.InvoiceReceivedEvent,
    models.InvoiceAcceptedEvent,
    models.InvoiceReceivedEvent,
    models.InvoiceFailedEvent,
    models.InvoiceSettledEvent,
    models.InvoiceCancelledEvent,
]

DebitNoteEvent = Union[
    models.DebitNoteReceivedEvent,
    models.DebitNoteAcceptedEvent,
    models.DebitNoteReceivedEvent,
    models.DebitNoteFailedEvent,
    models.DebitNoteSettledEvent,
    models.DebitNoteCancelledEvent,
]


class PaymentEventCollector(YagnaEventCollector, ABC):
    def __init__(self, node: "GolemNode"):
        self.node = node
        self.min_ts = datetime.now(timezone.utc)

    def _collect_events_kwargs(self) -> Dict:
        return {"after_timestamp": self.min_ts, "app_session_id": self.node.app_session_id}

    async def _process_event(self, event: Union[InvoiceEvent, DebitNoteEvent]) -> None:
        self.min_ts = max(event.event_date, self.min_ts)
        resource, parent_resource = await self._get_event_resources(event)
        resource.add_event(event)
        if resource._parent is None:
            parent_resource.add_child(resource)

    @abstractmethod
    async def _get_event_resources(self, event: Any) -> Tuple[Resource, Resource]:
        raise NotImplementedError



import datetime
import enum
import inspect
from dataclasses import Field
from typing import Any

from golem.utils.typing import match_type_union_aware


class PropsConsSerializerMixin:
    @classmethod
    def _serialize_value(cls, value: Any) -> Any:
        """Return value in primitive format compatible with Golem's property \
        and constraint syntax."""

        if isinstance(value, (list, tuple)):
            return type(value)(cls._serialize_value(v) for v in value)

        if isinstance(value, datetime.datetime):
            return int(value.replace(tzinfo=datetime.timezone.utc).timestamp() * 1000)

        if isinstance(value, enum.Enum):
            return value.value

        return value

    @classmethod
    def _deserialize_value(cls, value: Any, field: Field) -> Any:
        """Return proper value for field from given primitive.

        Intended to be overriden with additional type serialisation methods.
        """
        if matched_type := match_type_union_aware(
            field.type, lambda t: inspect.isclass(t) and issubclass(t, datetime.datetime)
        ):
            return matched_type.fromtimestamp(int(float(value) * 0.001), datetime.timezone.utc)

        if matched_type := match_type_union_aware(
            field.type, lambda t: inspect.isclass(t) and issubclass(t, enum.Enum)
        ):
            return matched_type(value)

        return value



import asyncio
import contextvars
import inspect
import logging
from datetime import timedelta
from typing import Coroutine, Iterable, Optional, Tuple, TypeVar, cast

from golem.utils.logging import trace_id_var
from golem.utils.typing import MaybeAwaitable

T = TypeVar("T")

logger = logging.getLogger(__name__)


def create_task_with_logging(coro: Coroutine, *, trace_id: Optional[str] = None) -> asyncio.Task:
    task, task_name = _create_task_with_context(coro, trace_id=trace_id)

    logger.debug("Task `%s` created", task_name)

    return task


def create_delayed_task_with_logging(
    delay: timedelta, coro: Coroutine, *, trace_id: Optional[str] = None
) -> asyncio.Task:
    task, task_name = _create_task_with_context(
        _create_delayed_task_with_logging(delay, coro), trace_id=trace_id
    )

    logger.debug("Task `%s` created with delay `%s`", task_name, delay)

    return task


async def _create_delayed_task_with_logging(delay: timedelta, coro: Coroutine):
    try:
        await asyncio.sleep(delay.total_seconds())
    except asyncio.CancelledError:
        coro.close()
        raise

    return await coro


def _create_task_with_context(
    coro: Coroutine, *, trace_id: Optional[str] = None
) -> Tuple[asyncio.Task, str]:
    context = contextvars.copy_context()
    task = context.run(_create_task_with_logging, coro, trace_id=trace_id)

    if trace_id is not None:
        task_name = trace_id
    else:
        task_name = task.get_name()

    return task, task_name


def _create_task_with_logging(coro: Coroutine, *, trace_id: Optional[str] = None) -> asyncio.Task:
    if trace_id is not None:
        trace_id_var.set(trace_id)

    task = asyncio.create_task(coro)
    task.add_done_callback(_handle_task_logging)
    return task


def _handle_task_logging(task: asyncio.Task):
    try:
        return task.result()
    except asyncio.CancelledError:
        pass
    except Exception:
        logger.exception("Background async task encountered unhandled exception!")


async def ensure_cancelled(task: asyncio.Task) -> None:
    """Cancel given task and await for its cancellation."""

    if task.done():
        return

    task.cancel()

    try:
        await task
    except asyncio.CancelledError:
        pass


async def ensure_cancelled_many(tasks: Iterable[asyncio.Task]) -> None:
    """Cancel given tasks and concurrently await for their cancellation."""

    await asyncio.gather(*[ensure_cancelled(task) for task in tasks])


async def resolve_maybe_awaitable(value: MaybeAwaitable[T]) -> T:
    """Return given value or await for it results if value is awaitable."""

    if inspect.isawaitable(value):
        return await value

    # TODO: remove cast as inspect.isawaitable can't tell mypy that at this
    #  point value is T
    return cast(T, value)



import asyncio
import inspect
from typing import AsyncIterator, Awaitable, Callable, Generic, Tuple, TypeVar, Union

from golem.pipeline.exceptions import InputStreamExhausted

InType = TypeVar("InType")
OutType = TypeVar("OutType")


async def default_on_exception(func: Callable, args: Tuple, orig_exc: Exception) -> None:
    args_str = ", ".join([str(arg) for arg in args])
    print(f"Exception in {func.__name__}({args_str}): {orig_exc}")


class Map(Generic[InType, OutType]):
    """Turns one async iterator into another async iterator using a provided mapping function.

    Sample usage::

        async def src_func():
            yield 1
            yield 2

        async def map_func(x):
            return x * 2

        async for awaitable_val in Map(map_func)(src_func()):
            print(await awaitable_val)
            #   2
            #   4

    Caveats:
    * Always yields awaitables
    * It doesn't matter if source iterator yields `X` or `Awaitable[X]`, which has two consequences:

       * (good) Maps can be stacked one after another in a :any:`Chain`
       * (bad) Mapping functions that accept an awaitable as an argument should be avoided.

    * If input stream yields tuples, they will be passed to mapping function unpacked

    """

    def __init__(
        self,
        func: Callable[[InType], Awaitable[OutType]],
        *,
        on_exception: Callable[
            [Callable, Tuple, Exception], Awaitable[None]
        ] = default_on_exception,
    ):
        """Init Map.

        :param func: An async function that will be executed on every element of the stream passed
            to :any:`__call__`.
        :param on_exception: An async function that will be executed whenever main function raises
            an exception.
            Defaults to a function that prints the exception.
        """
        self.func = func
        self.on_exception = on_exception

    async def __call__(
        self,
        in_stream: Union[AsyncIterator[InType], AsyncIterator[Awaitable[InType]]],
    ) -> AsyncIterator[Awaitable[OutType]]:
        """Call Map.

        :param in_stream: An async stream of either func args or args-returning awaitables.
        """
        self._in_stream_lock = asyncio.Lock()
        while True:
            yield asyncio.create_task(self._next_value(in_stream))

    async def _next_value(
        self, in_stream: Union[AsyncIterator[InType], AsyncIterator[Awaitable[InType]]]
    ) -> OutType:
        while True:
            async with self._in_stream_lock:
                try:
                    in_val = await in_stream.__anext__()
                except StopAsyncIteration:
                    raise InputStreamExhausted()

            args = await self._as_awaited_tuple(in_val)

            try:
                return await self.func(*args)
            except Exception as e:
                await self.on_exception(self.func, args, e)

    async def _as_awaited_tuple(self, in_val: Union[InType, Awaitable[InType], Tuple]) -> Tuple:
        #   Q: Why this?
        #   A: Because this way it's possible to wait chains of awaitables without
        #      dealing with awaitables at all. E.g. We have Map(X -> Y) followed by Map(Y -> Z)
        #      and first map returns Awaitable[Y], and second map unpacks this Awaitable here.
        #   (This probably has some downsides, but should be worth it)
        if not isinstance(in_val, tuple):
            if inspect.isawaitable(in_val):
                awaited_val = await in_val
                if isinstance(awaited_val, tuple):
                    return awaited_val
                else:
                    return (awaited_val,)
            else:
                return (in_val,)
        else:
            new_vals = []
            for single_val in in_val:
                if inspect.isawaitable(single_val):
                    single_val = await single_val
                new_vals.append(single_val)
            return tuple(new_vals)



from typing import TYPE_CHECKING

from golem.resources.exceptions import ResourceException

if TYPE_CHECKING:
    from golem.resources.network.network import Network


class NetworkException(ResourceException):
    pass


class NetworkFull(NetworkException):
    """Raised when we need a new free ip but there are no free ips left in the :any:`Network`."""

    def __init__(self, network: "Network"):
        self._network = network
        super().__init__(f"{network} is full - there are no free ips left")

    @property
    def network(self) -> "Network":
        return self._network



from golem.resources.network.events import NetworkClosed, NetworkDataChanged, NewNetwork
from golem.resources.network.exceptions import NetworkException, NetworkFull
from golem.resources.network.network import DeployArgsType, Network

__all__ = (
    "Network",
    "DeployArgsType",
    "NetworkException",
    "NetworkFull",
    "NewNetwork",
    "NetworkDataChanged",
    "NetworkClosed",
)



from golem.payload import defaults
from golem.payload.base import Payload, constraint, prop
from golem.payload.constraints import (
    Constraint,
    ConstraintException,
    ConstraintGroup,
    Constraints,
    PropertyName,
    PropertyValue,
)
from golem.payload.defaults import ActivityInfo, NodeInfo, PaymentInfo
from golem.payload.exceptions import InvalidProperties, PayloadException
from golem.payload.generic import GenericPayload
from golem.payload.parser import PayloadSyntaxParser, SyntaxException
from golem.payload.properties import Properties
from golem.payload.vm import ManifestVmPayload, RepositoryVmPayload, VmPayload, VmPayloadException

__all__ = (
    "Payload",
    "GenericPayload",
    "prop",
    "constraint",
    "VmPayload",
    "RepositoryVmPayload",
    "ManifestVmPayload",
    "VmPayloadException",
    "Constraints",
    "Constraint",
    "ConstraintGroup",
    "Properties",
    "PayloadException",
    "ConstraintException",
    "InvalidProperties",
    "SyntaxException",
    "PayloadSyntaxParser",
    "PropertyName",
    "PropertyValue",
    "defaults",
    "ActivityInfo",
    "NodeInfo",
    "PaymentInfo",
)



from golem.resources.invoice.event_collectors import InvoiceEventCollector
from golem.resources.invoice.events import InvoiceClosed, InvoiceDataChanged, NewInvoice
from golem.resources.invoice.invoice import Invoice

__all__ = (
    "Invoice",
    "InvoiceEventCollector",
    "NewInvoice",
    "InvoiceDataChanged",
    "InvoiceClosed",
)



from typing import List

from golem.payload.base import Constraints, Payload, Properties
from golem.payload.parser import PayloadSyntaxParser


class GenericPayload(Payload):
    """A generic demand specification implementing the `Payload` interface."""

    def __init__(self, properties: Properties, constraints: Constraints):
        self._properties: Properties = properties
        self._constraints: Constraints = constraints
        super().__init__()

    def _build_properties(self) -> Properties:
        return self._properties

    def _build_constraints(self) -> Constraints:
        return self._constraints

    @classmethod
    def from_raw_api_data(cls, properties: List, constraints: List[str]):
        return cls(
            properties=Properties({prop.key: prop.value for prop in properties}),
            constraints=Constraints(
                [PayloadSyntaxParser.get_instance().parse_constraints(c) for c in constraints]
            ),
        )



import asyncio
import logging
from collections import defaultdict
from dataclasses import dataclass
from typing import Awaitable, Callable, DefaultDict, List, Optional, Tuple, Type

from golem.event_bus.base import Event, EventBus, EventBusError, TEvent
from golem.utils.asyncio import create_task_with_logging, ensure_cancelled
from golem.utils.logging import get_trace_id_name, trace_span

logger = logging.getLogger(__name__)


@dataclass
class _CallbackInfo:
    callback: Callable[[TEvent], Awaitable[None]]
    filter_func: Optional[Callable[[TEvent], bool]]
    once: bool


_CallbackHandler = Tuple[Type[TEvent], _CallbackInfo]


class InMemoryEventBus(EventBus[_CallbackHandler]):
    def __init__(self):
        self._callbacks: DefaultDict[Type[Event], List[_CallbackInfo]] = defaultdict(list)
        self._event_queue: asyncio.Queue[Event] = asyncio.Queue()
        self._process_event_queue_loop_task: Optional[asyncio.Task] = None

    @trace_span()
    async def start(self):
        if self.is_started():
            raise EventBusError("Event bus is already started!")

        self._process_event_queue_loop_task = create_task_with_logging(
            self._process_event_queue_loop(),
            trace_id=get_trace_id_name(self, "process-event-queue-loop"),
        )

    @trace_span()
    async def stop(self):
        await self._event_queue.join()

        if self._process_event_queue_loop_task is not None:
            await ensure_cancelled(self._process_event_queue_loop_task)
            self._process_event_queue_loop_task = None

    @trace_span(show_results=True)
    def is_started(self) -> bool:
        return (
            self._process_event_queue_loop_task is not None
            and not self._process_event_queue_loop_task.done()
        )

    @trace_span(show_arguments=True)
    async def on(
        self,
        event_type: Type[TEvent],
        callback: Callable[[TEvent], Awaitable[None]],
        filter_func: Optional[Callable[[TEvent], bool]] = None,
    ) -> _CallbackHandler:
        callback_info = _CallbackInfo(
            callback=callback,
            filter_func=filter_func,
            once=False,
        )

        self._callbacks[event_type].append(callback_info)

        callback_handler = (event_type, callback_info)

        return callback_handler

    @trace_span(show_arguments=True)
    async def on_once(
        self,
        event_type: Type[TEvent],
        callback: Callable[[TEvent], Awaitable[None]],
        filter_func: Optional[Callable[[TEvent], bool]] = None,
    ) -> _CallbackHandler:
        callback_info = _CallbackInfo(
            callback=callback,
            filter_func=filter_func,
            once=True,
        )

        self._callbacks[event_type].append(callback_info)

        callback_handler = (event_type, callback_info)

        return callback_handler

    @trace_span(show_arguments=True)
    async def off(self, callback_handler: _CallbackHandler) -> None:
        event_type, callback_info = callback_handler
        try:
            self._callbacks[event_type].remove(callback_info)
        except (KeyError, ValueError):
            message = "Given callback handler is not found in event bus!"
            logger.debug(
                f"Removing callback handler `{id(callback_handler)}` failed with `{message}`"
            )
            raise EventBusError(message)

    @trace_span(show_arguments=True)
    async def emit(self, event: TEvent) -> None:
        if not self.is_started():
            message = "Event bus is not started!"
            logger.debug(f"Emitting event `{event}` failed with `message`")
            raise EventBusError(message)

        await self._event_queue.put(event)

    async def _process_event_queue_loop(self):
        while True:
            logger.debug("Getting event from queue...")
            event = await self._event_queue.get()
            logger.debug(f"Getting event from queue done with `{event}`")

            for event_type, callback_infos in self._callbacks.items():
                await self._process_event(event, event_type, callback_infos)

            self._event_queue.task_done()

    @trace_span(show_arguments=True)
    async def _process_event(
        self, event: Event, event_type: Type[Event], callback_infos: List[_CallbackInfo]
    ):
        if not isinstance(event, event_type):
            logger.debug(
                f"Processing event `{event}` on event type `{event_type}` ignored as event is"
                f" not a instance of event type"
            )
            return

        callback_infos_to_remove = []

        logger.debug(f"Processing callbacks for event {event}...")

        for callback_info in callback_infos:
            logger.debug(f"Processing callback {callback_info}...")

            if callback_info.filter_func is not None:
                logger.debug("Calling filter function...")
                try:
                    if not callback_info.filter_func(event):
                        logger.debug("Calling filter function done, ignoring callback")
                        continue
                except Exception:
                    logger.exception(
                        f"Encountered an error in `{callback_info.filter_func}` filter function"
                        f" while handling `{event}`!"
                    )
                    continue
                else:
                    logger.debug("Calling filter function done, calling callback")
            else:
                logger.debug("Callback has no filter function")

            logger.debug(f"Calling {callback_info.callback}...")
            try:
                # TODO: Support sync callbacks
                await callback_info.callback(event)
            except Exception as e:
                logger.debug(f"Calling {callback_info.callback} failed with `{e}")

                logger.exception(
                    f"Encountered an error in `{callback_info.callback}` callback"
                    f" while handling `{event}`!"
                )
                continue
            else:
                logger.debug(f"Calling {callback_info.callback} done")

            if callback_info.once:
                callback_infos_to_remove.append(callback_info)

                logger.debug(f"Callback {callback_info} marked as to be removed")

        logger.debug(f"Processing callbacks for event {event} done")

        if callback_infos_to_remove:
            logger.debug(f"Removing callbacks `{callback_infos_to_remove}`...")

            for callback_info in callback_infos_to_remove:
                callback_infos.remove(callback_info)

            logger.debug(f"Removing callbacks `{callback_infos_to_remove}` done")



from golem.resources.activity.activity import Activity
from golem.resources.activity.commands import Deploy, Start


# TODO: Move default functions to Activity class
async def default_prepare_activity(activity: Activity) -> Activity:
    """Execute Deploy() and Start() commands on a given :any:`Activity` and return the same \
    :any:`Activity`.

    If the commands fail, destroys the :any:`Activity` and terminates the corresponding
    :any:`Agreement`.
    """
    try:
        batch = await activity.execute_commands(Deploy(), Start())
        await batch.wait(timeout=300)
        assert batch.success, batch.events[-1].message
    except Exception:
        await activity.parent.close_all()
        raise
    return activity



import asyncio
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, AsyncIterator, Callable, Dict, List, Optional, Union, cast

from ya_market import RequestorApi
from ya_market import models as models

from golem.payload import Constraints, PayloadSyntaxParser, Properties
from golem.resources.base import _NULL, Resource, ResourceNotFound, api_call_wrapper
from golem.resources.demand.data import DemandData
from golem.resources.demand.events import DemandClosed, NewDemand
from golem.resources.proposal import Proposal
from golem.utils.low import YagnaEventCollector

if TYPE_CHECKING:
    from golem.node import GolemNode

DEFAULT_TTL = timedelta(hours=1)


class Demand(Resource[RequestorApi, models.Demand, _NULL, Proposal, _NULL], YagnaEventCollector):
    """A single demand on the Golem Network.

    Created with one of the :class:`Demand`-returning methods of the :any:`GolemNode`.
    """

    _demand_data: Optional[DemandData] = None

    def __init__(self, node: "GolemNode", id_: str, data: Optional[models.Demand] = None):
        super().__init__(node, id_, data)
        asyncio.create_task(node.event_bus.emit(NewDemand(self)))

    ######################
    #   EXTERNAL INTERFACE
    @api_call_wrapper(ignore_status_codes=(404, 410))
    async def unsubscribe(self) -> None:
        """Stop all operations related to this demand and remove it.

        This is a final operation, unsubscribed demand is not available anymore.
        """
        self.set_no_more_children()
        self.stop_collecting_events()
        await self.api.unsubscribe_demand(self.id)
        await self.node.event_bus.emit(DemandClosed(self))

    async def initial_proposals(self) -> AsyncIterator["Proposal"]:
        """Yield initial proposals matched to this demand."""
        async for proposal in self.child_aiter():
            assert isinstance(proposal, Proposal)  # mypy
            if proposal.initial:
                yield proposal

    def proposal(self, proposal_id: str) -> "Proposal":
        """Return a :class:`Proposal` with a given ID."""
        proposal = Proposal(self.node, proposal_id)

        #   NOTE: we don't know the parent, so we don't set it, but demand is known
        if proposal._demand is None and proposal._parent is None:
            proposal.demand = self

        return proposal

    ###########################
    #   Event collector methods
    def _collect_events_kwargs(self) -> Dict:
        return {"timeout": 5, "max_events": 10}

    def _collect_events_args(self) -> List:
        return [self.id]

    @property
    def _collect_events_func(self) -> Callable:
        return self.api.collect_offers

    async def _process_event(
        self, event: Union[models.ProposalEvent, models.ProposalRejectedEvent]
    ) -> None:
        if isinstance(event, models.ProposalEvent):
            proposal = Proposal.from_proposal_event(self.node, event)
            parent = self._get_proposal_parent(proposal)
            parent.add_child(proposal)
        elif isinstance(event, models.ProposalRejectedEvent):
            assert event.proposal_id is not None  # mypy
            proposal = self.proposal(event.proposal_id)
            proposal.add_event(event)

    #################
    #   OTHER METHODS
    @api_call_wrapper()
    async def _get_data(self) -> models.Demand:
        #   NOTE: this method is required because there is no get_demand(id)
        #         in ya_market (as there is no matching endpoint in yagna)
        all_demands: List[models.Demand] = await self.api.get_demands()
        try:
            return next(d for d in all_demands if d.demand_id == self.id)
        except StopIteration:
            raise ResourceNotFound(self)

    @classmethod
    async def create_from_properties_constraints(
        cls,
        node: "GolemNode",
        properties: Properties,
        constraints: Constraints,
    ) -> "Demand":
        data = models.DemandOfferBase(
            properties=properties.serialize(),
            constraints=constraints.serialize(),
        )
        return await cls.create(node, data)

    @classmethod
    async def create(cls, node: "GolemNode", data: models.DemandOfferBase) -> "Demand":
        api = cls._get_api(node)
        demand_id = await api.subscribe_demand(data)
        return cls(node, demand_id)

    def _get_proposal_parent(self, proposal: "Proposal") -> Union["Demand", "Proposal"]:
        assert proposal.data is not None

        if proposal.data.state == "Initial":
            parent = self
        else:
            parent_proposal_id = proposal.data.prev_proposal_id
            parent = Proposal(self.node, parent_proposal_id)  # type: ignore
        return parent

    async def get_demand_data(self) -> DemandData:
        if not self._demand_data:
            data = await self.get_data()

            # TODO: Make constraints parsing lazy
            constraints = PayloadSyntaxParser.get_instance().parse_constraints(data.constraints)

            return DemandData(
                properties=Properties(data.properties),
                constraints=constraints,
                demand_id=data.demand_id,
                requestor_id=data.requestor_id,
                timestamp=cast(datetime, data.timestamp),
            )

        return self._demand_data

    async def get_expiration_date(self) -> datetime:
        """Return expiration date to auto unsubscribe."""
        await self.get_data()

        return cast(datetime, self.data.timestamp) + DEFAULT_TTL



import asyncio
import logging
from typing import Awaitable, Callable, List

from golem.managers.base import Work, WorkManager, WorkResult
from golem.managers.work.mixins import WorkManagerDoWorkMixin, WorkManagerPluginsMixin
from golem.node import GolemNode
from golem.resources import Activity
from golem.utils.asyncio import create_task_with_logging
from golem.utils.logging import get_trace_id_name, trace_span

logger = logging.getLogger(__name__)


class ConcurrentWorkManager(WorkManagerPluginsMixin, WorkManagerDoWorkMixin, WorkManager):
    def __init__(
        self,
        golem: GolemNode,
        get_activity: Callable[[], Awaitable[Activity]],
        size: int,
        *args,
        **kwargs,
    ):
        self._get_activity = get_activity
        self._size = size

        self._queue: asyncio.Queue[Work] = asyncio.Queue()
        self._results: List[WorkResult] = []

        super().__init__(*args, **kwargs)

    @trace_span("Doing work", show_arguments=True, show_results=True, log_level=logging.INFO)
    async def do_work(self, work: Work) -> WorkResult:
        return await self._do_work_with_plugins(self._do_work, work)

    @trace_span("Doing work list", show_arguments=True, show_results=True, log_level=logging.INFO)
    async def do_work_list(self, work_list: List[Work]) -> List[WorkResult]:
        workers = [
            create_task_with_logging(self.worker(), trace_id=get_trace_id_name(self, f"worker-{i}"))
            for i in range(self._size)
        ]
        for work in work_list:
            self._queue.put_nowait(work)
        await self._queue.join()
        [w.cancel() for w in workers]
        await asyncio.gather(*workers, return_exceptions=True)
        return self._results

    async def worker(self):
        while True:
            work = await self._get_work_from_queue()

            result = await self.do_work(work)
            self._queue.task_done()
            self._results.append(result)

    @trace_span(show_results=True)
    async def _get_work_from_queue(self) -> Work:
        return await self._queue.get()



from abc import ABC
from typing import TYPE_CHECKING, Any, Dict, Generic, Tuple, TypeVar

from golem.event_bus import Event

if TYPE_CHECKING:
    from golem.resources.base import Resource  # noqa

TResourceEvent = TypeVar("TResourceEvent", bound="ResourceEvent")
TResource = TypeVar("TResource", bound="Resource")


class ResourceEvent(Event, ABC, Generic[TResource]):
    """Base class for all events related to a particular :any:`Resource`."""

    def __init__(self, resource: TResource):
        self._resource = resource

    @property
    def resource(self) -> TResource:
        """Resource related to this :class:`ResourceEvent`."""
        return self._resource

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.resource})"


class NewResource(ResourceEvent[TResource], ABC, Generic[TResource]):
    """Emitted when a new :any:`Resource` object is created.

    There are three distinct scenarios possible:

    * We create a new resource, e.g. with :any:`GolemNode.create_allocation()`
    * We start interacting with some resource that was created by us before,
      but not with this particular GolemNode instance, eg. :any:`GolemNode.allocation()`
    * We find a resource created by someone else (e.g. a :any:`Proposal` or a :any:`DebitNote`)

    There's no difference between these scenarios from the POV of this event.
    """


class ResourceDataChanged(ResourceEvent[TResource], Generic[TResource]):
    """Emitted when `data` attribute of a :any:`Resource` changes.

    This event is **not** emitted when the data "would have changed if we
    requested new data, but we didn't request it". In other words, we don't listen for
    `yagna`-side changes, only react to changes already noticed in the context of a
    :any:`GolemNode`.

    Second argument is the old data (before the change), so comparing
    `event.resource.data` with `event.old_data` shows what changed.

    NULL (i.e. empty) change doesn't trigger the change, even if we explicitly sent a
    resource-changing call.
    """

    def __init__(self, resource: TResource, old_data: Any):
        super().__init__(resource)
        self._old_data = old_data

        #   TODO: this should be included in typing, but I don't know how to do this
        assert old_data is None or type(resource.data) is type(old_data)

    @property
    def old_data(self) -> Any:
        """Value of `self.resource.data` before the change."""
        return self._old_data

    def diff(self) -> Dict[str, Tuple[Any, Any]]:
        """Return a dictionary {property_name: (old_val, new_val)} with all values that changed."""
        old_dict = self.old_data.to_dict()
        new_dict = self.resource.data.to_dict()
        diff_dict = {}

        for key in old_dict.keys():
            old_val = old_dict[key]
            new_val = new_dict[key]
            if old_val != new_val:
                diff_dict[key] = (old_val, new_val)

        return diff_dict

    def __repr__(self) -> str:
        diff = []
        for key, (old_val, new_val) in self.diff().items():
            diff.append(f"{key}: {old_val} -> {new_val}")
        diff_str = ", ".join(diff)
        return f"{type(self).__name__}({self.resource}, {diff_str})"


class ResourceClosed(ResourceEvent[TResource], Generic[TResource]):
    """Emitted when a resource is deleted or rendered unusable.

    Usual case is when we delete a resource (e.g. :any:`Allocation.release()`),
    or when the lifespan of a resource ends (e.g. :any:`Agreement.terminate()`),
    but this can be also emitted when we notice that resource was deleted by someone
    else (TODO: when? is this possible at all? e.g. expired proposal?).

    This is emitted only when something changes. Creating a new Resource object
    for an already closed resource (e.g. by passing an id of a terminated agreement to
    :any:`GolemNode.agreement`) does not trigger this event.

    This event should never be emitted more than once for a given :any:`Resource`.
    Currently this is not true for :any:`Activity` - this is a known TODO
    (https://github.com/golemfactory/golem-core-python/issues/33).

    Not all resources are closed, e.g. :any:`PoolingBatch` is not.
    """



import logging
from datetime import datetime
from typing import cast

from golem.managers.base import PricingCallable, ProposalManagerPlugin
from golem.payload import Constraints, Properties
from golem.resources import Proposal, ProposalData
from golem.utils.logging import trace_span

logger = logging.getLogger(__name__)


class RejectIfCostsExceeds(ProposalManagerPlugin):
    def __init__(
        self, cost: float, pricing_callable: PricingCallable, reject_on_unpriceable=True
    ) -> None:
        self._cost = cost
        self._pricing_callable = pricing_callable
        self._reject_on_unpriceable = reject_on_unpriceable

    @trace_span(show_results=True)
    async def get_proposal(self) -> Proposal:
        while True:
            proposal: Proposal = await self._get_proposal()
            proposal_data = await self._get_proposal_data_from_proposal(proposal)

            cost = self._pricing_callable(proposal_data)

            if cost is None and self._reject_on_unpriceable:
                if not proposal.initial:
                    await proposal.reject("Can't estimate costs!")

                logger.debug(
                    "Can't estimate proposal `%s` costs `%s`, picking different one...",
                    proposal,
                    self._pricing_callable,
                )

                continue

            if cost is not None and self._cost < cost:
                if not proposal.initial:
                    await proposal.reject(
                        f"Exceeds costs `{self._pricing_callable}` limit of `{self._cost}`!"
                    )

                logger.debug(
                    "Proposal `%s` costs `%s` of `%f` exceeds limit of `%f`,"
                    " picking different one...",
                    proposal,
                    self._pricing_callable,
                    cost,
                    self._cost,
                )

                continue

            return proposal

    async def _get_proposal_data_from_proposal(self, proposal: Proposal) -> ProposalData:
        data = await proposal.get_data()

        return ProposalData(
            properties=Properties(data.properties),
            constraints=Constraints(),
            proposal_id=data.proposal_id,
            issuer_id=data.issuer_id,
            state=data.state,
            timestamp=cast(datetime, data.timestamp),
            prev_proposal_id=data.prev_proposal_id,
        )



from golem.managers.activity import PoolActivityManager, SingleUseActivityManager
from golem.managers.agreement import DefaultAgreementManager
from golem.managers.base import (
    ActivityManager,
    AgreementManager,
    DemandManager,
    DoWorkCallable,
    Manager,
    NetworkManager,
    PaymentManager,
    ProposalManager,
    ProposalManagerPlugin,
    ProposalScorer,
    RejectProposal,
    Work,
    WorkContext,
    WorkManager,
    WorkResult,
)
from golem.managers.demand import AggregatingDemandManager, RefreshingDemandManager
from golem.managers.mixins import BackgroundLoopMixin
from golem.managers.network import SingleNetworkManager
from golem.managers.payment import DefaultPaymentManager
from golem.managers.proposal import (
    BlacklistProviderIdPlugin,
    DefaultProposalManager,
    LinearAverageCostPricing,
    LinearCoeffsCost,
    LinearPerCpuAverageCostPricing,
    LinearPerCpuCoeffsCost,
    MapScore,
    MidAgreementPaymentsNegotiator,
    NegotiatingPlugin,
    PaymentPlatformNegotiator,
    PropertyValueLerpScore,
    ProposalBuffer,
    ProposalScoringBuffer,
    ProposalScoringMixin,
    RandomScore,
    RejectIfCostsExceeds,
)
from golem.managers.work import (
    ConcurrentWorkManager,
    SequentialWorkManager,
    WorkManagerPluginsMixin,
    redundancy_cancel_others_on_first_done,
    retry,
    work_plugin,
)

__all__ = (
    "PoolActivityManager",
    "SingleUseActivityManager",
    "DefaultAgreementManager",
    "DoWorkCallable",
    "Manager",
    "ProposalScorer",
    "RejectProposal",
    "Work",
    "WorkManager",
    "WorkContext",
    "WorkResult",
    "NetworkManager",
    "PaymentManager",
    "ProposalManager",
    "ProposalManagerPlugin",
    "AggregatingDemandManager",
    "DemandManager",
    "AgreementManager",
    "ActivityManager",
    "RefreshingDemandManager",
    "BackgroundLoopMixin",
    "SingleNetworkManager",
    "DefaultPaymentManager",
    "DefaultProposalManager",
    "BlacklistProviderIdPlugin",
    "ProposalBuffer",
    "PaymentPlatformNegotiator",
    "MidAgreementPaymentsNegotiator",
    "NegotiatingPlugin",
    "RejectIfCostsExceeds",
    "MapScore",
    "ProposalScoringMixin",
    "LinearAverageCostPricing",
    "LinearPerCpuAverageCostPricing",
    "LinearPerCpuCoeffsCost",
    "LinearCoeffsCost",
    "PropertyValueLerpScore",
    "RandomScore",
    "ProposalScoringBuffer",
    "SequentialWorkManager",
    "ConcurrentWorkManager",
    "WorkManagerPluginsMixin",
    "redundancy_cancel_others_on_first_done",
    "retry",
    "work_plugin",
)



from golem.utils.low.api import ActivityApi, ApiConfig, ApiFactory, TRequestorApi, get_requestor_api
from golem.utils.low.event_collector import YagnaEventCollector

__all__ = (
    "YagnaEventCollector",
    "TRequestorApi",
    "ActivityApi",
    "ApiConfig",
    "ApiFactory",
    "get_requestor_api",
)
